"""Modbus client async UDP communication."""
from __future__ import annotations

import socket
import time
from collections.abc import Callable

from pymodbus.client.base import ModbusBaseClient, ModbusBaseSyncClient
from pymodbus.exceptions import ConnectionException
from pymodbus.framer import FramerType
from pymodbus.logging import Log
from pymodbus.pdu import ModbusPDU
from pymodbus.transport import CommParams, CommType


DGRAM_TYPE = socket.SOCK_DGRAM


class AsyncModbusUdpClient(ModbusBaseClient):
    """**AsyncModbusUdpClient**.

    Fixed parameters:

    :param host: Host IP address or host name

    Optional parameters:

    :param framer: Framer name, default FramerType.SOCKET
    :param port: Port used for communication.
    :param name: Set communication name, used in logging
    :param source_address: source address of client,
    :param reconnect_delay: Minimum delay in seconds.milliseconds before reconnecting.
    :param reconnect_delay_max: Maximum delay in seconds.milliseconds before reconnecting.
    :param timeout: Timeout for connecting and receiving data, in seconds.
    :param retries: Max number of retries per request.
    :param trace_packet: Called with bytestream received/to be sent
    :param trace_pdu: Called with PDU received/to be sent
    :param trace_connect: Called when connected/disconnected

    .. tip::
        The trace methods allow to modify the datastream/pdu !

    .. tip::
        **reconnect_delay** doubles automatically with each unsuccessful connect, from
        **reconnect_delay** to **reconnect_delay_max**.
        Set `reconnect_delay=0` to avoid automatic reconnection.

    Example::

        from pymodbus.client import AsyncModbusUdpClient

        async def run():
            client = AsyncModbusUdpClient("localhost")

            await client.connect()
            ...
            client.close()

    Please refer to :ref:`Pymodbus internals` for advanced usage.
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        host: str,
        *,
        framer: FramerType = FramerType.SOCKET,
        port: int = 502,
        name: str = "comm",
        source_address: tuple[str, int] | None = None,
        reconnect_delay: float = 0.1,
        reconnect_delay_max: float = 300,
        timeout: float = 3,
        retries: int = 3,
        trace_packet: Callable[[bool, bytes], bytes] | None = None,
        trace_pdu: Callable[[bool, ModbusPDU], ModbusPDU] | None = None,
        trace_connect: Callable[[bool], None] | None = None,
    ) -> None:
        """Initialize Asyncio Modbus UDP Client."""
        self.comm_params = CommParams(
            comm_type=CommType.UDP,
            host=host,
            port=port,
            comm_name=name,
            source_address=source_address,
            reconnect_delay=reconnect_delay,
            reconnect_delay_max=reconnect_delay_max,
            timeout_connect=timeout,
        )
        if framer not in [FramerType.SOCKET, FramerType.RTU, FramerType.ASCII]:
            raise TypeError("Only FramerType SOCKET/RTU/ASCII allowed.")
        ModbusBaseClient.__init__(
            self,
            framer,
            retries,
            self.comm_params,
            trace_packet,
            trace_pdu,
            trace_connect,
        )
        self.source_address = source_address


class ModbusUdpClient(ModbusBaseSyncClient):
    """**ModbusUdpClient**.

    Fixed parameters:

    :param host: Host IP address or host name

    Optional parameters:

    :param framer: Framer name, default FramerType.SOCKET
    :param port: Port used for communication.
    :param name: Set communication name, used in logging
    :param source_address: source address of client,
    :param reconnect_delay: Not used in the sync client
    :param reconnect_delay_max: Not used in the sync client
    :param timeout: Timeout for connecting and receiving data, in seconds.
    :param retries: Max number of retries per request.
    :param trace_packet: Called with bytestream received/to be sent
    :param trace_pdu: Called with PDU received/to be sent
    :param trace_connect: Called when connected/disconnected

    .. tip::
        The trace methods allow to modify the datastream/pdu !

    Example::

        from pymodbus.client import ModbusUdpClient

        async def run():
            client = ModbusUdpClient("localhost")

            client.connect()
            ...
            client.close()

    Please refer to :ref:`Pymodbus internals` for advanced usage.
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        host: str,
        *,
        framer: FramerType = FramerType.SOCKET,
        port: int = 502,
        name: str = "comm",
        source_address: tuple[str, int] | None = None,
        reconnect_delay: float = 0.1,
        reconnect_delay_max: float = 300,
        timeout: float = 3,
        retries: int = 3,
        trace_packet: Callable[[bool, bytes], bytes] | None = None,
        trace_pdu: Callable[[bool, ModbusPDU], ModbusPDU] | None = None,
        trace_connect: Callable[[bool], None] | None = None,
    ) -> None:
        """Initialize Modbus UDP Client."""
        if framer not in [FramerType.SOCKET, FramerType.RTU, FramerType.ASCII]:
            raise TypeError("Only FramerType SOCKET/RTU/ASCII allowed.")
        self.comm_params = CommParams(
            comm_type=CommType.UDP,
            host=host,
            port=port,
            comm_name=name,
            source_address=source_address,
            reconnect_delay=reconnect_delay,
            reconnect_delay_max=reconnect_delay_max,
            timeout_connect=timeout,
        )
        super().__init__(
            framer,
            retries,
            self.comm_params,
            trace_packet,
            trace_pdu,
            trace_connect,
        )
        self.socket: socket.socket | None = None

    @property
    def connected(self) -> bool:
        """Connect internal."""
        return self.socket is not None

    def connect(self):
        """Connect to the modbus udp server.

        :meta private:
        """
        if self.socket:
            return True
        try:
            self.socket = socket.socket(-1, socket.SOCK_DGRAM)
            self.socket.settimeout(self.comm_params.timeout_connect)
        except OSError as exc:
            Log.error("Unable to create udp socket {}", exc)
            self.close()
        return self.socket is not None

    def close(self):
        """Close the underlying socket connection.

        :meta private:
        """
        self.socket = None

    def send(self, request: bytes, addr: tuple | None = None) -> int:
        """Send data on the underlying socket.

        :meta private:
        """
        _ = addr
        if not self.socket:
            raise ConnectionException(str(self))
        if request:
            return self.socket.sendto(
                request, (self.comm_params.host, self.comm_params.port)
            )
        return 0

    def recv(self, size: int | None) -> bytes:
        """Read data from the underlying descriptor.

        :meta private:
        """
        if not self.socket:
            raise ConnectionException(str(self))
        if size is None:
            size = 4096
        data = self.socket.recvfrom(size)[0]
        self.last_frame_end = round(time.time(), 6)
        return data

    def is_socket_open(self):
        """Check if socket is open.

        :meta private:
        """
        return True

    def __repr__(self):
        """Return string representation."""
        return (
            f"<{self.__class__.__name__} at {hex(id(self))} socket={self.socket}, "
            f"ipaddr={self.comm_params.host}, port={self.comm_params.port}, timeout={self.comm_params.timeout_connect}>"
        )



"""Pymodbus: Modbus Protocol Implementation.

Released under the BSD license
"""

__all__ = [
    "ExceptionResponse",
    "FramerType",
    "ModbusException",
    "__version__",
    "__version_full__",
    "pymodbus_apply_logging_config",
]

from pymodbus.exceptions import ModbusException
from pymodbus.framer import FramerType
from pymodbus.logging import pymodbus_apply_logging_config
from pymodbus.pdu import ExceptionResponse


__version__ = "3.9.2"
__version_full__ = f"[pymodbus, version {__version__}]"



"""Wrappers to call pyproject.toml-based build backend hooks.
"""

from typing import TYPE_CHECKING

from ._impl import (
    BackendUnavailable,
    BuildBackendHookCaller,
    HookMissing,
    UnsupportedOperation,
    default_subprocess_runner,
    quiet_subprocess_runner,
)

__version__ = "1.2.0"
__all__ = [
    "BackendUnavailable",
    "BackendInvalid",
    "HookMissing",
    "UnsupportedOperation",
    "default_subprocess_runner",
    "quiet_subprocess_runner",
    "BuildBackendHookCaller",
]

BackendInvalid = BackendUnavailable  # Deprecated alias, previously a separate exception

if TYPE_CHECKING:
    from ._impl import SubprocessRunner

    __all__ += ["SubprocessRunner"]



# Copyright 2009-2017 Wander Lairson Costa
# Copyright 2009-2021 PyUSB contributors
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# All the hacks necessary to assure compatibility across all
# supported versions come here.
# Please, note that there is one version check for each
# hack we need to do, this makes maintenance easier... ^^

import sys
import array
import usb.util as util

__all__ = []

# we support Python >= 3.9
assert sys.hexversion >= 0x030900f0

def as_array(data_or_length=None):
    """Convert loosely specified `data_or_length` to a byte array.

    The following types of `data_or_length` are supported:

    - an `array('B')` value (no-op, returns it back);
    - the `None` value (returns an new empty array);
    - an integer length (returns a new array with the specified size);
    - lists or iterables of small enough integers;
    - strings.
    """

    if isinstance(data_or_length, array.array) and data_or_length.typecode == 'B':
        return data_or_length

    if isinstance(data_or_length, str):
        return array.array('B', data_or_length.encode('utf-8'))

    if data_or_length is None:
        return array.array('B')

    try:
        return util.create_buffer(data_or_length)
    except TypeError:
        return array.array('B', data_or_length)



"""Base for all clients."""
from __future__ import annotations

import asyncio
from abc import abstractmethod
from collections.abc import Awaitable, Callable

from pymodbus.client.mixin import ModbusClientMixin
from pymodbus.exceptions import ConnectionException
from pymodbus.framer import FRAMER_NAME_TO_CLASS, FramerBase, FramerType
from pymodbus.logging import Log
from pymodbus.pdu import DecodePDU, ModbusPDU
from pymodbus.transaction import TransactionManager
from pymodbus.transport import CommParams


class ModbusBaseClient(ModbusClientMixin[Awaitable[ModbusPDU]]):
    """**ModbusBaseClient**.

    :mod:`ModbusBaseClient` is normally not referenced outside :mod:`pymodbus`.
    """

    def __init__(
        self,
        framer: FramerType,
        retries: int,
        comm_params: CommParams,
        trace_packet: Callable[[bool, bytes], bytes] | None,
        trace_pdu: Callable[[bool, ModbusPDU], ModbusPDU] | None,
        trace_connect: Callable[[bool], None] | None,
    ) -> None:
        """Initialize a client instance.

        :meta private:
        """
        ModbusClientMixin.__init__(self)  # type: ignore[arg-type]
        self.comm_params = comm_params
        self.ctx = TransactionManager(
            comm_params,
            (FRAMER_NAME_TO_CLASS[framer])(DecodePDU(False)),
            retries,
            False,
            trace_packet,
            trace_pdu,
            trace_connect,
        )

    @property
    def connected(self) -> bool:
        """Return state of connection."""
        return self.ctx.is_active()

    async def connect(self) -> bool:
        """Call transport connect."""
        self.ctx.reset_delay()
        Log.debug(
            "Connecting to {}:{}.",
            self.ctx.comm_params.host,
            self.ctx.comm_params.port,
        )
        rc = await self.ctx.connect()
        await asyncio.sleep(0.1)
        return rc

    def register(self, custom_response_class: type[ModbusPDU]) -> None:
        """Register a custom response class with the decoder (call **sync**).

        :param custom_response_class: (optional) Modbus response class.
        :raises MessageRegisterException: Check exception text.

        Use register() to add non-standard responses (like e.g. a login prompt) and
        have them interpreted automatically.
        """
        self.ctx.framer.decoder.register(custom_response_class)

    def close(self) -> None:
        """Close connection."""
        self.ctx.close()

    def execute(self, no_response_expected: bool, request: ModbusPDU):
        """Execute request and get response (call **sync/async**).

        :meta private:
        """
        if not self.ctx.transport:
            raise ConnectionException(f"Not connected[{self!s}]")
        return self.ctx.execute(no_response_expected, request)

    def set_max_no_responses(self, max_count: int) -> None:
        """Override default max no request responses.

        :param max_count: Max aborted requests before disconnecting.

        The parameter retries defines how many times a request is retried
        before being aborted. Once aborted a counter is incremented, and when
        this counter is greater than max_count the connection is terminated.

        .. tip::
            When a request is successful the count is reset.
        """
        self.ctx.max_until_disconnect = max_count

    async def __aenter__(self):
        """Implement the client with enter block.

        :returns: The current instance of the client
        :raises ConnectionException:
        """
        await self.connect()
        return self

    async def __aexit__(self, klass, value, traceback):
        """Implement the client with aexit block."""
        self.close()

    def __str__(self):
        """Build a string representation of the connection.

        :returns: The string representation
        """
        return (
            f"{self.__class__.__name__} {self.ctx.comm_params.host}:{self.ctx.comm_params.port}"
        )


class ModbusBaseSyncClient(ModbusClientMixin[ModbusPDU]):
    """**ModbusBaseClient**.

    :mod:`ModbusBaseClient` is normally not referenced outside :mod:`pymodbus`.
    """

    def __init__(
        self,
        framer: FramerType,
        retries: int,
        comm_params: CommParams,
        trace_packet: Callable[[bool, bytes], bytes] | None,
        trace_pdu: Callable[[bool, ModbusPDU], ModbusPDU] | None,
        trace_connect: Callable[[bool], None] | None,
    ) -> None:
        """Initialize a client instance.

        :meta private:
        """
        ModbusClientMixin.__init__(self)  # type: ignore[arg-type]
        self.comm_params = comm_params
        self.retries = retries
        self.slaves: list[int] = []

        # Common variables.
        self.framer: FramerBase = (FRAMER_NAME_TO_CLASS[framer])(DecodePDU(False))
        self.transaction = TransactionManager(
            self.comm_params,
            self.framer,
            retries,
            False,
            trace_packet,
            trace_pdu,
            trace_connect,
            sync_client=self,
        )
        self.reconnect_delay_current = self.comm_params.reconnect_delay or 0
        self.use_udp = False
        self.last_frame_end: float | None = 0
        self.silent_interval: float = 0

    # ----------------------------------------------------------------------- #
    # Client external interface
    # ----------------------------------------------------------------------- #
    def register(self, custom_response_class: type[ModbusPDU]) -> None:
        """Register a custom response class with the decoder.

        :param custom_response_class: (optional) Modbus response class.
        :raises MessageRegisterException: Check exception text.

        Use register() to add non-standard responses (like e.g. a login prompt) and
        have them interpreted automatically.
        """
        self.framer.decoder.register(custom_response_class)

    def idle_time(self) -> float:
        """Time before initiating next transaction (call **sync**).

        Applications can call message functions without checking idle_time(),
        this is done automatically.
        """
        if self.last_frame_end is None or self.silent_interval is None:
            return 0
        return self.last_frame_end + self.silent_interval

    def execute(self, no_response_expected: bool, request: ModbusPDU) -> ModbusPDU:
        """Execute request and get response (call **sync/async**).

        :param no_response_expected: The client will not expect a response to the request
        :param request: The request to process
        :returns: The result of the request execution
        :raises ConnectionException: Check exception text.

        :meta private:
        """
        if not self.connect():
            raise ConnectionException(f"Failed to connect[{self!s}]")
        return self.transaction.sync_execute(no_response_expected, request)

    def set_max_no_responses(self, max_count: int) -> None:
        """Override default max no request responses.

        :param max_count: Max aborted requests before disconnecting.

        The parameter retries defines how many times a request is retried
        before being aborted. Once aborted a counter is incremented, and when
        this counter is greater than max_count the connection is terminated.

        .. tip::
            When a request is successful the count is reset.
        """
        self.transaction.max_until_disconnect = max_count

    # ----------------------------------------------------------------------- #
    # Internal methods
    # ----------------------------------------------------------------------- #
    @abstractmethod
    def send(self, request: bytes, addr: tuple | None = None) -> int:
        """Send request.

        :meta private:
        """

    @abstractmethod
    def recv(self, size: int | None) -> bytes:
        """Receive data.

        :meta private:
        """

    def connect(self) -> bool:  # type: ignore[empty-body]
        """Connect to other end, overwritten."""

    def close(self):
        """Close connection, overwritten."""

    # ----------------------------------------------------------------------- #
    # The magic methods
    # ----------------------------------------------------------------------- #
    def __enter__(self):
        """Implement the client with enter block.

        :returns: The current instance of the client
        :raises ConnectionException:
        """
        self.connect()
        return self

    def __exit__(self, klass, value, traceback):
        """Implement the client with exit block."""
        self.close()

    def __str__(self):
        """Build a string representation of the connection.

        :returns: The string representation
        """
        return (
            f"{self.__class__.__name__} {self.comm_params.host}:{self.comm_params.port}"
        )



# -*- coding: utf-8 -*-

"""scinstr.vaccum
author    Benoit Dubois
copyright FEMTO ENGINEERING, 2024
license   GPL v3.0+
brief     Class to handle Edwards D395 controller device.
details   All messages consist of ASCII coded characters. Messages to the ADC
start with either a "!" or a "?" character. All messages end with a carriage
return (cr). Characters not enclosed by start (!?) and end (cr) characters
will be ignored. Incomplete messages will be ignored if a new start character
is received. Responses from the ADC end with a carriage return (cr).
If the message cannot be understood or if the syntax is wrong, an error message
of the form "Err n" will be returned.
"""

import logging
import serial
import scinstr.vacuum.gauge as gauge


TIMEOUT = 0.5

CR = '\x0D'

COMMANDS = [
    '!CH', # Set-point high threshold
    '!CL', # Set-point low threshold
    '!GA', # Accept gauge error
    '!GW', # Switch gauge on/off
    '!RC', # Relay controlling gauge
    '!TH', # Link high threshold
    '!TL', # Link low threshold
    '!US', # Units
]

QUERIES = [
    '?CH', # Set-point high threshold -> {pressure}
    '?CL', # Set-point low threshold -> {pressure}
    '?GA', # Gauge pressure -> {pressure}
    '?GV', # Gauge version -> {nn}
    '?RC', # Relay controlling gauge -> {1..2}
    '?TH', # Link high threshold -> {pressure}
    '?TL', # Link low threshold -> {pressure}
    '?US', # Units -> {0..3}
    '?VL', # Voltage -> {voltage}
]

ERROR_CODES = {
    # Controller errors
    '0': 'No error',
    '1': 'EEPROM error',
    '2': 'ID reference error',
    # Gauge errors
    '11': 'Gauge voltage too high',
    '12': 'Gauge voltage too low',
    '13': 'AIM Gauge not striking',
    '21': 'WRG Pirani failure',
    '22': 'WRG magnetron short',
    '23': 'WRG striker fail',
    '24': 'WRG magnetron not struck',
    '25': 'APGX filament failure',
    '26': 'APGX cal err',
    '27': 'APGXH tube not fitted',
    # RS232 errors
    '51': 'Not a valid query or command word',
    '52': 'Message incomplete',
    '53': 'Message too long',
    '54': 'Incorrect gauge number',
    '57': 'Incorrect number format',
    '58': 'Incorrect pressure format',
    #
    '81': 'No gauge connected',
    '82': 'Unknown gauge type',
    '83': 'Gauge not reading pressure',
    '84': 'AIM gauge striking',
    '90': 'Incorrect gauge type, query/commandnot appropriate',
    '91': 'Gauge turn-on is inhibited by link',
}


def errn_to_errmsg(errn):
    print("ERROR:", retval)
    print(retval[-1])
    error_message = ERROR_CODES[retval[-1]]
    print(error_message)


class CommunicationError(Exception):
    """Raised when device return error message"""
    def __init__(self, errn):
        msg = errn_to_errmsg(errn)
        super().__init__(msg)



# =============================================================================
class D395:
    """Handle vacuum controller device through RS232-serial/USB connection.
    The USB interface emulates a 'classic' serial interface.
    Default RS232 protocol: 9600 baud, 1 stop bit, 8 data bits, no parity.
    """

    def __init__(self, port=None, baudrate=9600, bytesize=8,
                 parity=serial.PARITY_NONE, stopbits=1, timeout=TIMEOUT):
        super().__init__()
        self._ser = serial.Serial()
        self.port = port
        self.baudrate = baudrate
        self.bytesize = bytesize
        self.parity = parity
        self.stopbits = stopbits
        self.timeout = timeout

    def connect(self):
        """Connect to the remote host.
        :returns: True if connection succeeded, False otherwise.
        """
        if self._ser.isOpen() is False:
            try:
                self._ser.open()
            except ValueError as ex:
                logging.error("Wrong connection parameters: %r", ex)
                return False
            except serial.SerialTimeoutException:
                logging.error("Timeout on connection")
                return False
            except serial.serialutil.SerialException as ex:
                logging.error("Serial error: %r", ex)
                return False
        logging.info("Connected to gauge")
        self._ser.flush()
        return True

    def close(self):
        """Close the underlying serial connection.
        """
        if self._ser is not None:
            self._ser.close()
        self._ser = None
        logging.info("Connection to gauge closed")

    def _write(self, data):
        """Specific device writing process.
        :param data: data writes to device (str)
        :returns: number of bytes sent (int)
        """
        nb_bytes = self._ser.write(data.encode('utf-8'))
        return nb_bytes

    def _read(self, length):
        """Specific device reading process.
        :param length: length of message to read (int)
        :returns: Message reads from device (str)
        """
        retval = self._ser.read(length).decode('utf-8')
        return retval

    def read(self, length=128):
        """Basic read process.
        :param length: max number of bytes to read, None -> no limit (int)
        :returns: data received or False if communication error (str)
        """
        retval = self._read(length)
        retval = retval.strip(CR)
        logging.debug("read %r", retval)
        return retval

    def command(self, mnemo, parameters=None):
        """Command message.
        :param mnemo: mnemonique (command mnemonique) (str)
        :param parameters: command parameters if needed (str)
        :returns: True if transmission OK else False (bool)
        """
        assert mnemo in COMMANDS
        if parameters is not None:
            msg = mnemo + '=' + str(parameters) + CR
        else:
            msg = mnemo + CR
        self._write(msg)
        retval = self._read(8).strip(CR)
        if retval != 'Err0':
            logging.debug("Command %r error: %r", msg.strip(CR), errmsg)
            raise CommunicationError(retval)
        logging.debug("Command %r ok", msg.strip(CR))
        return True

    def query(self, mnemo, parameter=None, size=100):
        """Ask procedure ie read data after write.
        :param mnemo: mnemonique (command/response) (str)
        :param parameters: query parameters if needed (str)
        :param size: size of returned data (int)
        :returns: data received (str)
        """
        assert mnemo in QUERIES
        if parameter is not None:
            msg = mnemo + str(parameter) + CR
        else:
            msg = mnemo + CR
        self._write(msg)
        retval = self._read(128).strip(CR)
        if 'Err' in retval:
            logging.debug("Querry %r error: %r", msg.strip(CR))
            raise CommunicationError(retval)
        return retval

    def pressure(self, gauge_nb):
        """Ask pressure value of gauge 'gauge_nb'.
        :param gauge_nb: id number of the gauge (int)
        :returns: pressure value (float)
        """
        return self.query('?GA', gauge_nb)

    @property
    def port(self):
        return self._ser.port

    @port.setter
    def port(self, value):
        self._ser.port = value

    @property
    def baudrate(self):
        return self._ser.baudrate

    @baudrate.setter
    def baudrate(self, value):
        self._ser.baudrate = value

    @property
    def bytesize(self):
        return self._ser.bytesize

    @bytesize.setter
    def bytesize(self, value):
        self._ser.bytesize = value

    @property
    def parity(self):
        return self._ser.parity

    @parity.setter
    def parity(self, value):
        self._ser.parity = value

    @property
    def stopbits(self):
        return self._ser.stopbits

    @stopbits.setter
    def stopbits(self, value):
        self._ser.stopbits = value

    @property
    def timeout(self):
        return self._ser.timeout

    @timeout.setter
    def timeout(self, value):
        self._ser.timeout = value




# =============================================================================
if __name__ == '__main__':
    import logging
    import argparse

    PARSER = argparse.ArgumentParser(description='Test Edwards D395')
    PARSER.add_argument('-p', '--port', dest='port', type=str,
                        help='Serial port to connect to')

    ARGS = PARSER.parse_args()
    PORT = ARGS.port

    INST = D395(port=PORT)
    INST.connect()
    print("Pressure:", INST.pressure(1))
    INST.close()



# -*- coding: utf-8 -*-

"""package t7pro
author    Labajack, Benoit Dubois
copyright Labjack
brief     Migrate to Python3 from Labjack source.
"""

import struct


def unpack_helper(fmt, data):
    size = struct.calcsize(fmt)
    return struct.unpack(fmt, data[:size]), data[size:]


def hex2float(s):
    """Function modified to handle Python 3.
    """
    if s[0:2] == '0x':
        s = s[2:len(s)]
    bins = bytes.fromhex(s)
    return struct.unpack('>f', bins)[0]


def float2int(num):
    return struct.unpack('=i', struct.pack('=f', num))[0]


def concatData(data):
    tVal = 0
    upper = True
    for reg in data:
        if upper:
            tVal = ((reg & 0xFFFF) << 16)
            upper = False
        else:
            tVal = tVal | (reg & 0xFFFF)
            upper = True
    return tVal


'''
Converting Numbers to 16bit data array's
'''


def uint16_to_data(num):
    return struct.unpack('=H', struct.pack('=H', num & 0xFFFF))[0]


def uint32_to_data(num):
    data = [0, 0]
    data[0] = struct.unpack('=H', struct.pack('=H', (num >> 16) & 0xffff))[0]
    data[1] = struct.unpack('=H', struct.pack('=H', num & 0xffff))[0]
    return data


def int32_to_data(num):
    data = [0, 0]
    data[0] = struct.unpack('=H', struct.pack('=H', (num >> 16) & 0xffff))[0]
    data[1] = struct.unpack('=H', struct.pack('=H', num & 0xffff))[0]
    return data


def float32_to_data(num):
    intNum = float2int(num)
    data = [0, 0]
    data[0] = (intNum >> 16) & 0xFFFF
    data[1] = intNum & 0xFFFF
    return data


'''
Converting Data array's to Numbers
'''


def data_to_uint16(data):
    return data[0]


def data_to_uint32(data):
    return concatData(data)


def data_to_int32(data):
    return struct.unpack('=i', struct.pack('=I', concatData(data)))[0]


def data_to_float32(data):
    return hex2float(hex(concatData(data)))



"""Modbus Utilities.

A collection of utilities for packing data, unpacking
data computing checksums, and decode checksums.
"""
from __future__ import annotations


# pylint: disable=missing-type-doc


def dict_property(store, index):
    """Create class properties from a dictionary.

    Basically this allows you to remove a lot of possible
    boilerplate code.

    :param store: The store store to pull from
    :param index: The index into the store to close over
    :returns: An initialized property set
    """
    if hasattr(store, "__call__"):
        getter = lambda self: store(  # pylint: disable=unnecessary-lambda-assignment
            self
        )[index]
        setter = lambda self, value: store(  # pylint: disable=unnecessary-lambda-assignment
            self
        ).__setitem__(index, value)
    elif isinstance(store, str):
        getter = lambda self: self.__getattribute__(  # pylint: disable=unnecessary-dunder-call,unnecessary-lambda-assignment
            store
        )[index]
        setter = lambda self, value: self.__getattribute__(  # pylint: disable=unnecessary-dunder-call,unnecessary-lambda-assignment
            store
        ).__setitem__(index, value)
    else:
        getter = (
            lambda self: store[index]  # pylint: disable=unnecessary-lambda-assignment
        )
        setter = lambda self, value: store.__setitem__(  # pylint: disable=unnecessary-lambda-assignment
            index, value
        )

    return property(getter, setter)


# --------------------------------------------------------------------------- #
# Error Detection Functions
# --------------------------------------------------------------------------- #


def hexlify_packets(packet):
    """Return hex representation of bytestring received.

    :param packet:
    :return:
    """
    if not packet:
        return ""
    return " ".join([hex(int(x)) for x in packet])



"""Diagnostic Record Read/Write."""
from __future__ import annotations

import struct
from typing import cast

from pymodbus.constants import ModbusPlusOperation
from pymodbus.datastore import ModbusSlaveContext
from pymodbus.device import ModbusControlBlock
from pymodbus.pdu.pdu import ModbusPDU, pack_bitstring


_MCB = ModbusControlBlock()


class DiagnosticBase(ModbusPDU):
    """DiagnosticBase."""

    function_code = 0x08
    sub_function_code: int = 9999
    rtu_frame_size = 8

    def __init__(self, message: bytes | int | list | tuple | None = 0, dev_id: int = 1, transaction_id: int = 0) -> None:
        """Initialize a diagnostic request."""
        super().__init__(transaction_id=transaction_id, dev_id=dev_id)
        self.message: bytes | int | list | tuple | None = message

    def encode(self) -> bytes:
        """Encode a diagnostic response."""
        packet = struct.pack(">H", self.sub_function_code)
        if self.message is not None:
            if isinstance(self.message, bytes):
                packet += self.message
                return packet
            if isinstance(self.message, int):
                packet += struct.pack(">H", self.message)
                return packet
            if isinstance(self.message, (list, tuple)):
                for piece in self.message:
                    packet += struct.pack(">H", piece)
                return packet
            raise TypeError(f"UNKNOWN DIAG message type: {type(self.message)}")
        return packet

    def decode(self, data: bytes) -> None:
        """Decode a diagnostic request."""
        (self.sub_function_code, ) = struct.unpack(">H", data[:2])
        data = data[2:]
        if self.sub_function_code == ReturnQueryDataRequest.sub_function_code:
            self.message = data
        elif (data_len := len(data)):
            if data_len % 2:
                data_len += 1
                data += b"0"
            if (word_len := data_len // 2) == 1:
                (self.message,) = struct.unpack(">H", data)
            else:
                self.message = struct.unpack(">" + "H" * word_len, data)

    def get_response_pdu_size(self) -> int:
        """Get response pdu size.

        Func_code (1 byte) + Sub function code (2 byte) + Data (2 * N bytes)
        """
        return 1 + 2 + 2

    async def update_datastore(self, _context: ModbusSlaveContext) -> ModbusPDU:
        """Implement dummy."""
        response = {
            DiagnosticBase.sub_function_code: DiagnosticBase,
            ReturnQueryDataResponse.sub_function_code: ReturnQueryDataResponse,
            RestartCommunicationsOptionResponse.sub_function_code: RestartCommunicationsOptionResponse,
        }[self.sub_function_code]
        return response(message=self.message, dev_id=self.dev_id, transaction_id=self.transaction_id)


class ReturnQueryDataRequest(DiagnosticBase):
    """ReturnQueryDataRequest."""

    sub_function_code = 0x0000


class ReturnQueryDataResponse(DiagnosticBase):
    """ReturnQueryDataResponse."""

    sub_function_code = 0x0000


class RestartCommunicationsOptionRequest(DiagnosticBase):
    """RestartCommunicationsOptionRequest."""

    sub_function_code = 0x0001


class RestartCommunicationsOptionResponse(DiagnosticBase):
    """RestartCommunicationsOptionResponse."""

    sub_function_code = 0x0001


class ReturnDiagnosticRegisterRequest(DiagnosticBase):
    """ReturnDiagnosticRegisterRequest."""

    sub_function_code = 0x0002

    async def update_datastore(self, _context: ModbusSlaveContext) -> ModbusPDU:
        """update_datastore the diagnostic request on the given device."""
        register = pack_bitstring(_MCB.getDiagnosticRegister())
        return ReturnDiagnosticRegisterResponse(message=register, dev_id=self.dev_id, transaction_id=self.transaction_id)


class ReturnDiagnosticRegisterResponse(DiagnosticBase):
    """ReturnDiagnosticRegisterResponse."""

    sub_function_code = 0x0002


class ChangeAsciiInputDelimiterRequest(DiagnosticBase):
    """ChangeAsciiInputDelimiterRequest."""

    sub_function_code = 0x0003

    async def update_datastore(self, _context: ModbusSlaveContext) -> ModbusPDU:
        """update_datastore the diagnostic request on the given device."""
        char = (cast(int, self.message) & 0xFF00) >> 8
        _MCB.Delimiter = char
        return ChangeAsciiInputDelimiterResponse(message=self.message, dev_id=self.dev_id, transaction_id=self.transaction_id)


class ChangeAsciiInputDelimiterResponse(DiagnosticBase):
    """ChangeAsciiInputDelimiterResponse."""

    sub_function_code = 0x0003


class ForceListenOnlyModeRequest(DiagnosticBase):
    """ForceListenOnlyModeRequest."""

    sub_function_code = 0x0004

    async def update_datastore(self, _context: ModbusSlaveContext) -> ModbusPDU:
        """update_datastore the diagnostic request on the given device."""
        _MCB.ListenOnly = True
        return ForceListenOnlyModeResponse(dev_id=self.dev_id, transaction_id=self.transaction_id)


class ForceListenOnlyModeResponse(DiagnosticBase):
    """ForceListenOnlyModeResponse.

    This does not send a response
    """

    sub_function_code = 0x0004

    def __init__(self, dev_id=1, transaction_id=0):
        """Initialize to block a return response."""
        DiagnosticBase.__init__(self, dev_id=dev_id, transaction_id=transaction_id)
        self.message = []


class ClearCountersRequest(DiagnosticBase):
    """ClearCountersRequest."""

    sub_function_code = 0x000A

    async def update_datastore(self, _context: ModbusSlaveContext) -> ModbusPDU:
        """update_datastore the diagnostic request on the given device."""
        _MCB.reset()
        return ClearCountersResponse(dev_id=self.dev_id, transaction_id=self.transaction_id)


class ClearCountersResponse(DiagnosticBase):
    """ClearCountersResponse."""

    sub_function_code = 0x000A


class ReturnBusMessageCountRequest(DiagnosticBase):
    """ReturnBusMessageCountRequest."""

    sub_function_code = 0x000B

    async def update_datastore(self, _context: ModbusSlaveContext) -> ModbusPDU:
        """update_datastore the diagnostic request on the given device."""
        count = _MCB.Counter.BusMessage
        return ReturnBusMessageCountResponse(message=count, dev_id=self.dev_id, transaction_id=self.transaction_id)


class ReturnBusMessageCountResponse(DiagnosticBase):
    """ReturnBusMessageCountResponse."""

    sub_function_code = 0x000B


class ReturnBusCommunicationErrorCountRequest(DiagnosticBase):
    """ReturnBusCommunicationErrorCountRequest."""

    sub_function_code = 0x000C

    async def update_datastore(self, _context: ModbusSlaveContext) -> ModbusPDU:
        """update_datastore the diagnostic request on the given device."""
        count = _MCB.Counter.BusCommunicationError
        return ReturnBusCommunicationErrorCountResponse(message=count, dev_id=self.dev_id, transaction_id=self.transaction_id)


class ReturnBusCommunicationErrorCountResponse(DiagnosticBase):
    """ReturnBusCommunicationErrorCountResponse."""

    sub_function_code = 0x000C


class ReturnBusExceptionErrorCountRequest(DiagnosticBase):
    """ReturnBusExceptionErrorCountRequest."""

    sub_function_code = 0x000D

    async def update_datastore(self, _context: ModbusSlaveContext) -> ModbusPDU:
        """update_datastore the diagnostic request on the given device."""
        count = _MCB.Counter.BusExceptionError
        return ReturnBusExceptionErrorCountResponse(message=count, dev_id=self.dev_id, transaction_id=self.transaction_id)


class ReturnBusExceptionErrorCountResponse(DiagnosticBase):
    """ReturnBusExceptionErrorCountResponse."""

    sub_function_code = 0x000D


class ReturnSlaveMessageCountRequest(DiagnosticBase):
    """ReturnSlaveMessageCountRequest."""

    sub_function_code = 0x000E

    async def update_datastore(self, _context: ModbusSlaveContext) -> ModbusPDU:
        """update_datastore the diagnostic request on the given device."""
        count = _MCB.Counter.SlaveMessage
        return ReturnSlaveMessageCountResponse(message=count, dev_id=self.dev_id, transaction_id=self.transaction_id)


class ReturnSlaveMessageCountResponse(DiagnosticBase):
    """ReturnSlaveMessageCountResponse."""

    sub_function_code = 0x000E


class ReturnSlaveNoResponseCountRequest(DiagnosticBase):
    """ReturnSlaveNoResponseCountRequest."""

    sub_function_code = 0x000F

    async def update_datastore(self, _context: ModbusSlaveContext) -> ModbusPDU:
        """update_datastore the diagnostic request on the given device."""
        count = _MCB.Counter.SlaveNoResponse
        return ReturnSlaveNoResponseCountResponse(message=count, dev_id=self.dev_id, transaction_id=self.transaction_id)


class ReturnSlaveNoResponseCountResponse(DiagnosticBase):
    """ReturnSlaveNoResponseCountResponse."""

    sub_function_code = 0x000F


class ReturnSlaveNAKCountRequest(DiagnosticBase):
    """ReturnSlaveNAKCountRequest."""

    sub_function_code = 0x0010

    async def update_datastore(self, _context: ModbusSlaveContext) -> ModbusPDU:
        """update_datastore the diagnostic request on the given device."""
        count = _MCB.Counter.SlaveNAK
        return ReturnSlaveNAKCountResponse(message=count, dev_id=self.dev_id, transaction_id=self.transaction_id)


class ReturnSlaveNAKCountResponse(DiagnosticBase):
    """ReturnSlaveNAKCountResponse."""

    sub_function_code = 0x0010


class ReturnSlaveBusyCountRequest(DiagnosticBase):
    """ReturnSlaveBusyCountRequest."""

    sub_function_code = 0x0011

    async def update_datastore(self, _context: ModbusSlaveContext) -> ModbusPDU:
        """update_datastore the diagnostic request on the given device."""
        count = _MCB.Counter.SLAVE_BUSY
        return ReturnSlaveBusyCountResponse(message=count, dev_id=self.dev_id, transaction_id=self.transaction_id)


class ReturnSlaveBusyCountResponse(DiagnosticBase):
    """ReturnSlaveBusyCountResponse."""

    sub_function_code = 0x0011


class ReturnSlaveBusCharacterOverrunCountRequest(DiagnosticBase):
    """ReturnSlaveBusCharacterOverrunCountRequest."""

    sub_function_code = 0x0012

    async def update_datastore(self, _context: ModbusSlaveContext) -> ModbusPDU:
        """update_datastore the diagnostic request on the given device."""
        count = _MCB.Counter.BusCharacterOverrun
        return ReturnSlaveBusCharacterOverrunCountResponse(message=count, dev_id=self.dev_id, transaction_id=self.transaction_id)


class ReturnSlaveBusCharacterOverrunCountResponse(DiagnosticBase):
    """ReturnSlaveBusCharacterOverrunCountResponse."""

    sub_function_code = 0x0012


class ReturnIopOverrunCountRequest(DiagnosticBase):
    """ReturnIopOverrunCountRequest."""

    sub_function_code = 0x0013

    async def update_datastore(self, _context: ModbusSlaveContext) -> ModbusPDU:
        """update_datastore the diagnostic request on the given device."""
        count = _MCB.Counter.BusCharacterOverrun
        return ReturnIopOverrunCountResponse(message=count, dev_id=self.dev_id, transaction_id=self.transaction_id)


class ReturnIopOverrunCountResponse(DiagnosticBase):
    """ReturnIopOverrunCountResponse."""

    sub_function_code = 0x0013


class ClearOverrunCountRequest(DiagnosticBase):
    """ClearOverrunCountRequest."""

    sub_function_code = 0x0014

    async def update_datastore(self, _context: ModbusSlaveContext) -> ModbusPDU:
        """update_datastore the diagnostic request on the given device."""
        _MCB.Counter.BusCharacterOverrun = 0x0000
        return ClearOverrunCountResponse(dev_id=self.dev_id, transaction_id=self.transaction_id)


class ClearOverrunCountResponse(DiagnosticBase):
    """ClearOverrunCountResponse."""

    sub_function_code = 0x0014


class GetClearModbusPlusRequest(DiagnosticBase):
    """GetClearModbusPlusRequest."""

    sub_function_code = 0x0015

    def get_response_pdu_size(self):
        """Return size of the respaonse.

        Func_code (1 byte) + Sub function code (2 byte) + Operation (2 byte) + Data (108 bytes)
        """
        data = 2 + 108 if self.message == ModbusPlusOperation.GET_STATISTICS else 0
        return 1 + 2 + 2 + 2 + data

    async def update_datastore(self, _context: ModbusSlaveContext) -> ModbusPDU:
        """update_datastore the diagnostic request on the given device."""
        message: int | list | None = None  # the clear operation does not return info
        if self.message == ModbusPlusOperation.CLEAR_STATISTICS:
            _MCB.Plus.reset()
            message = self.message
        else:
            message = [self.message]
            message += _MCB.Plus.encode()
        return GetClearModbusPlusResponse(message=message, dev_id=self.dev_id, transaction_id=self.transaction_id)

    def encode(self):
        """Encode a diagnostic response."""
        packet = struct.pack(">H", self.sub_function_code)
        packet += struct.pack(">H", self.message)
        return packet


class GetClearModbusPlusResponse(DiagnosticBase):
    """GetClearModbusPlusResponse."""

    sub_function_code = 0x0015



# -*- coding: utf-8 -*-

"""package scinstr
author    Benoit Dubois
copyright FEMTO ENGINEERING, 2019
license   GPL v3.0+
brief     API dedicated to handle the Keysight 34461A digital multimeter.
"""

import logging
import time
import re
import socket
import usbtmc

from scinstr.dmm.constants import DMM34461A_PORT, DMM34461A_TIMEOUT, \
    DMM34461A_ID, DMM34461A_VID, DMM34461A_PID

# Multimeter functions
(CAP, CURR, CURRDC, VOLTAC, VOLTDC, VOLTDCRAT, RES, FRES, FRE, PER,
 TEMPRTD, TEMPFRTD, DIOD, CONT) = ("CAP", "CURR", "CURR:DC", "VOLT:AC",
                                   "VOLT:DC", "VOLT:DC:RAT", "RES",
                                   "FRES", "FRE", "PER", "TEMP:RTD",
                                   "TEMP:FRTD", "DIOD", "CONT")

# Integration time
(PLC002, PLC02, PLC1, PLC10, PLC100) = ("0.02", "0.2", "1", "10", "100")

# Voltage input range
(RANGE10MV, RANGE1V, RANGE10V, RANGE100V, RANGE1000V) = ("0.1", "1", "10",
                                                         "100", "1000")

# Trigger source
(IMM, EXT, BUS) = ("IMM", "EXT", "BUS")


# =============================================================================
class Dmm34461aAbstract():
    """Abstract class to handling DMM34461a digital multimeter device. Derived
    classes need to re-implement specific-protocol methods: connect(), close(),
    _write(), _read()...
    """

    def __del__(self):
        if self.is_connected:
            self.local()

    def connect(self):
        """Abstract protocol connect process. Derived classes must implement
        the connect process dedicated to the specific protocol used.
        :returns: None
        """
        raise NotImplementedError("Method not implemented by derived class")

    @property
    def is_connected(self):
        """Abstract connection check property.
        :returns: True if device connected else False (bool)
        """
        NotImplementedError("Method not implemented by derived class")

    def close(self):
        """Abstract protocol closing process. Derived classes must implement
        the closing process dedicated to the specific protocol used.
        :returns: None
        """
        self.local()

    def _write(self, data):
        """Abstract protocol write process. Derived classes must implement
        the write process dedicated to the specific protocol used.
        :param data: data writes to device (str)
        :returns: None
        """
        NotImplementedError("Method not implemented by derived class")

    def _read(self, length):
        """Abstract protocol read process. Derived classes must implement
        the read process dedicated to the specific protocol used.
        :param length: length of message to read (int)
        :returns: Message reads from device (str)
        """
        NotImplementedError("Method not implemented by derived class")

    def write(self, data):
        """A basic write method: writes "data" to device.
        :param data: data writes to device (str)
        :returns: None
        """
        try:
            nb_bytes = self._write(data)
        except Exception as ex:
            logging.error("Write error: %r", ex)
            return 0
        logging.debug("write: %r", data)
        return nb_bytes

    def read(self, length):
        """A basic read method: read a message from device.
        :param length: length of message to read (int)
        :returns: Message reads from device (str)
        """
        try:
            retval = self._read(length)
        except Exception as ex:
            logging.error("Read error: %r", ex)
            return ''
        logging.debug("read: %r", retval)
        return retval

    def query(self, data, length=100):
        """Read after write method.
        :param data: data writes to device (str)
        :param length: length of message to read (int)
        :returns: Message returned by device (str)
        """
        self.write(data)
        return self.read(length)

    def ask(self, data, length=100):
        return self.query(data, length)

    def reset(self):
        """Resets meter to its power-on state, sets all bits to zero in
        status byte register and all event registers and clear error queue.
        :returns: None
        """
        try:
            self.write("*RST")
            time.sleep(2.0)
            self.write("*CLS")
            time.sleep(2.0)
        except Exception as ex:
            logging.error("reset failed %r", ex)
            return False
        logging.info("Reset meter")
        return True

    def idn(self):
        """Return product id of device.
        :returns: product id of device (str)
        """
        self.write("*IDN?")
        _id = self.read(100)
        return _id

    def get_error(self):
        """Return list of current error.
        :returns: list of current error (list of str)
        """
        errors = [self.query("SYST:ERR?", 100)]
        while errors[-1] != "+0,\"No error\"":
            errors.append(self.query("SYST:ERR?", 100))
        return errors

    def check_interface(self):
        """Basic interface connection test: check id of device.
        Return True if interface with device is OK.
        :returns: status of interface with device (bool)
        """
        retval = self.connect()
        if retval is False:
            return False
        try:
            _id = str(self.idn())
        except Exception:  # Catch connection and idn() problem
            _id = ""  # To avoid error with 'find()' if '_id' is not defined
        self.close()
        if _id.find(DMM34461A_ID) >= 0:
            return True
        return False

    def local(self):
        self.query("SYST:LOC:REL")

    def dc_voltage_config(self, intg_time=None, rang=None, res=None,
                          d_fltr=None, a_fltr=None):
        """Configures specific parameters of DC voltage measurement.
        Note that you must specify a range whenever specifying a resolution
        parameter.
        :param intg_time: integration time in PLC (Power Line Cycle) (int)
        :param rang: DC voltage range (str)
        :param res: DC measurement resolution (int)
        :param d_fltr: Set digital averaging filter (bool)
        :param a_fltr: Set analog filter (bool)
        :returns: None
        """
        msg = "CONF:VOLT:DC"
        if rang is not None:
            msg += " " + rang
            if res is not None:
                msg += ", " + str(res)
        self.write(msg)
        if a_fltr is not None:
            self.write("FILT:DC:STAT" + str(a_fltr))
        if intg_time < 1:
            self.write("FILT:DC:DIG:STAT OFF")
        elif d_fltr is not None:
            self.write("FILT:DC:DIG:STAT" + str(d_fltr))
        if intg_time is not None:
            self.write("VOLT:DC:NPLC " + str(intg_time))
        logging.info("DC voltage measurement config done")

    def trigger_config(self, src=IMM, delay=None, delauto=None, scount=None,
                       tcount=None):
        """Configures trigger.
        :param src: source from which the Meter will expect a measurement
        trigger (str)
        :param delay: delay between receiving a trigger and the beginning of
        measurement cycle (int)
        :param delauto: enables or disables automatic trigger delay.
        The amount of delay is based on selected function, integration time,
        and filter setting. When automatic delay is enabled, "delay" parameters
        is ignored (int)
        :param scount: number of measurements the Meter takes per trigger
        (int or str)
        :param tcount: number of triggers the Meter will take before switching
        to an idle state. If you use "INF", continously accepts triggers,
        in this case use a device clear to stop the process (int or str)
        :returns: None
        """
        self.write("TRIG:SOUR " + src)
        if delauto == 1:
            self.write("TRIG:DEL:AUTO ON")
        elif delauto == 0:
            self.write("TRIG:DEL:AUTO OFF")
        elif delay is not None:
            self.write("TRIG:DEL " + str(delay))
        if tcount is not None:
            self.write("TRIG:COUN " + str(tcount))
        if scount is not None:
            self.write("SAMP:COUN " + str(scount))
        logging.info("Trigger configuration done")

    def init_trig(self):
        """Once the Meter has been configured for a measurement, ini_meas()
        causes the Meter to take a measurement when the trigger conditions have
        been met. The measurement reading(s) are placed in the Meter’s internal
        memory (up to 5000 to be read at a later time using the FETCh? command.
        :returns: None
        """
        self.write("INIT")
        logging.info("Initiate triggering")

    def data_read_fetch(self):
        """To process readings from the Meter’s internal memory to the output
        buffer.
        :returns: data read in internal memory buffer (float)
        """
        self.write("FETC?")
        retval = self.read(100)
        if re.match(r"[\+\-]\d\.\d*E[\+\-]\d\d", retval) is None:
            return None
        return float(retval)

    def data_read(self):
        """Takes a measurement the next time the trigger condition is met.
        After the measurement is taken, the reading is placed in the output
        buffer. "data_read" will not cause readings to be stored in the Meter’s
        internal memory.
        Read method convenient for slow measurement.
        :returns: data read in buffer (float)
        """
        self.write("READ?")
        retval = self.read(100)  # Note '100' not used in serial com mode
        if re.match(r"[\+\-]\d\.\d*E[\+\-]\d\d", retval) is None:
            logging.error("Invalid data %r", retval)
            return None
        return float(retval)


# =============================================================================
class Dmm34461aEth(Dmm34461aAbstract):
    """Handle DMM device through ethernet connection.
    """

    DEFAULT_PORT = DMM34461A_PORT
    DEFAULT_TIMEOUT = DMM34461A_TIMEOUT

    def __init__(self, ip=None, port=DMM34461A_PORT,
                 timeout=DMM34461A_TIMEOUT):
        """The constructor.
        :param ip: IP address of device (str)
        :param port: Ethernet port of device (int)
        :param timeout: Socket timeout value in s (float)
        :returns: None
        """
        self._sock = socket.socket(socket.AF_INET,
                                   socket.SOCK_STREAM,
                                   socket.IPPROTO_TCP)
        self._sock.settimeout(timeout)
        self.ip = ip
        self.port = port
        self._timeout = timeout

    def connect(self):
        """Specific ethernet connection process to DMM.
        :returns: True if connection success other False (Bool)
        """
        try:
            self._sock.connect((self.ip, self.port))
        except ValueError as ex:
            logging.error("Connection parameters out of range: %r", ex)
            return False
        except socket.timeout:
            logging.error("Timeout on connection")
            return False
        except Exception as ex:
            logging.critical("Connection with DMM problem: %r", ex)
            return False
        logging.info("Connected to DMM")
        return True

    @property
    def is_connected(self):
        if self._sock is None:
            return False
        return True

    def close(self):
        """Specific ethernet closing process with DMM.
        :returns: None
        """
        super().close()
        if self._sock is not None:
            self._sock.close()
        logging.info("Connection to DMM closed")

    def _write(self, data):
        """Specific ethernet writing process.
        :param data: data writes to device (str)
        :returns: number of byte sent (int)
        """
        return self._sock.send(data + '\n').encode('utf-8')

    def _read(self, length):
        """Specific ethernet reading process.
        :param length: length of message to read (int)
        :returns: message reads from device (str)
        """
        return self._sock.recv(length).decode('utf-8').strip('\n')

    @property
    def timeout(self):
        """Gets timeout on socket operations.
        :returns: timeout value in second (float)
        """
        return self._sock.gettimeout()

    @timeout.setter
    def timeout(self, timeout):
        """Sets timeout on socket operations.
        :param timeout: timeout value in second (float)
        :returns: None
        """
        self._sock.settimeout(timeout)


# =============================================================================
class Dmm34461aUsb(Dmm34461aAbstract):
    """Handle DMM device through USB connection.
    """

    def __init__(self,
                 vendor_id=DMM34461A_VID,
                 product_id=DMM34461A_PID,
                 timeout=DMM34461A_TIMEOUT):
        self.vid = vendor_id
        self.pid = product_id
        self._timeout = timeout
        self._dev = None

    def connect(self):
        """Connect to the remote host.
        :returns: True if connection succeeded, False otherwise
        """
        logging.info('Connecting to DMM')
        try:
            self._dev = usbtmc.Instrument(self.vid, self.pid)
        except Exception as ex:
            logging.error("USBTMC problem: %r", ex)
            return False
        self._dev.timeout = self._timeout
        try:
            self._dev.open()
        except Exception as ex:
            logging.error("Connection problem: %r", ex)
            return False
        logging.info('Connection --> Ok')
        return True

    @property
    def is_connected(self):
        if self._dev is None:
            return False
        return self._dev.connected

    def close(self):
        """Closes the underlying serial connection
        """
        super().close()
        try:
            self._dev.close()
        except Exception as ex:
            logging.error("Error when closing USB connection: %r", ex)
        logging.info("Connection to DMM closed")

    def _write(self, data):
        """Specific USB writing process.
        :param data: data writes to device (str)
        :returns: number of bytes sent (int)
        """
        try:
            nb_byte = self._dev.write(data)
        except AttributeError as ex:  # Exception raised after first write
            logging.debug(ex)         # No explication found
            nb_byte = self._dev.write(data)
        return nb_byte

    def _read(self, length):
        """Specific USB reading process.
        :param length: length of message to read (int)
        :returns: Message reads from device (str)
        """
        return self._dev.read(length)

    def ask(self, msg):
        return self._dev.ask(msg)

    @property
    def timeout(self):
        """Get timeout on socket operations.
        :returns: timeout value in second (float)
        """
        return self._timeout

    @timeout.setter
    def timeout(self, timeout):
        """Set timeout on socket operations.
        :param timeout: timeout value in second (float)
        :returns: None
        """
        self._timeout = timeout
        if self._dev is not None:
            self._dev.timeout = timeout


# =============================================================================
def check_dmm():
    """Check the Dmm34461axx class: connect to the multimeter, configure a dc
    voltage measurement then collect and print data to standard output.
    """
    from datetime import datetime

    date_fmt = "%d/%m/%Y %H:%M:%S"
    log_format = "%(asctime)s %(levelname) -8s %(filename)s " + \
                 " %(funcName)s (%(lineno)d): %(message)s"
    logging.basicConfig(level=logging.INFO,
                        datefmt=date_fmt,
                        format=log_format)

    dmm = Dmm34461aUsb(0x2a8d, 0x1601, timeout=4.8)
    # dmm = DmmEth(ip="192.168.0.61", port=5025, timeout=2.8)
    if dmm.connect() is not True:
        print("Connection failed")
        return
    dmm.reset()

    print("IDN:", dmm.query("*IDN?"))
    dmm.write("CONF:VOLT:DC AUTO")  # Autorange
    dmm.write("VOLT:DC:NPLC 100")
    dmm.write("VOLT:DC:ZERO:AUTO ON")  # Autozero off
    print("Error config?:", dmm.get_error())

    try:
        while True:
            value, errors = dmm.data_read()
            now = datetime.utcnow()
            if value is None or value == "":
                print("# No data @", now)
            else:
                print(now, value)
    except KeyboardInterrupt:
        dmm.write("ABORT")
    except Exception as er:
        logging.error("# Exception during acquisition: %r", er)

    print("Final error?:", dmm.get_error())

    dmm.close()


# =============================================================================
if __name__ == '__main__':
    check_dmm()



"""Modbus client async TLS communication."""
from __future__ import annotations

import socket
import ssl
from collections.abc import Callable

from pymodbus.client.tcp import AsyncModbusTcpClient, ModbusTcpClient
from pymodbus.framer import FramerType
from pymodbus.logging import Log
from pymodbus.pdu import ModbusPDU
from pymodbus.transport import CommParams, CommType


class AsyncModbusTlsClient(AsyncModbusTcpClient):
    """**AsyncModbusTlsClient**.

    Fixed parameters:

    :param host: Host IP address or host name

    Optional parameters:

    :param sslctx: SSLContext to use for TLS
    :param framer: Framer name, default FramerType.TLS
    :param port: Port used for communication
    :param name: Set communication name, used in logging
    :param source_address: Source address of client
    :param reconnect_delay: Minimum delay in seconds.milliseconds before reconnecting.
    :param reconnect_delay_max: Maximum delay in seconds.milliseconds before reconnecting.
    :param timeout: Timeout for connecting and receiving data, in seconds.
    :param retries: Max number of retries per request.
    :param trace_packet: Called with bytestream received/to be sent
    :param trace_pdu: Called with PDU received/to be sent
    :param trace_connect: Called when connected/disconnected

    .. tip::
        The trace methods allow to modify the datastream/pdu !

    .. tip::
        **reconnect_delay** doubles automatically with each unsuccessful connect, from
        **reconnect_delay** to **reconnect_delay_max**.
        Set `reconnect_delay=0` to avoid automatic reconnection.

    Example::

        from pymodbus.client import AsyncModbusTlsClient

        async def run():
            client = AsyncModbusTlsClient("localhost")

            await client.connect()
            ...
            client.close()

    Please refer to :ref:`Pymodbus internals` for advanced usage.
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        host: str,
        *,
        sslctx: ssl.SSLContext = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT),
        framer: FramerType = FramerType.TLS,
        port: int = 802,
        name: str = "comm",
        source_address: tuple[str, int] | None = None,
        reconnect_delay: float = 0.1,
        reconnect_delay_max: float = 300,
        timeout: float = 3,
        retries: int = 3,
        trace_packet: Callable[[bool, bytes], bytes] | None = None,
        trace_pdu: Callable[[bool, ModbusPDU], ModbusPDU] | None = None,
        trace_connect: Callable[[bool], None] | None = None,
    ):
        """Initialize Asyncio Modbus TLS Client."""
        if framer not in [FramerType.TLS]:
            raise TypeError("Only FramerType TLS allowed.")
        self.comm_params = CommParams(
            comm_type=CommType.TLS,
            host=host,
            sslctx=sslctx,
            port=port,
            comm_name=name,
            source_address=source_address,
            reconnect_delay=reconnect_delay,
            reconnect_delay_max=reconnect_delay_max,
            timeout_connect=timeout,
        )
        AsyncModbusTcpClient.__init__(
            self,
            "",
            framer=framer,
            retries=retries,
            trace_packet=trace_packet,
            trace_pdu=trace_pdu,
            trace_connect=trace_connect,
        )

    @classmethod
    def generate_ssl(
        cls,
        certfile: str | None = None,
        keyfile: str | None = None,
        password: str | None = None,
    ) -> ssl.SSLContext:
        """Generate sslctx from cert/key/password.

        :param certfile: Cert file path for TLS server request
        :param keyfile: Key file path for TLS server request
        :param password: Password for for decrypting private key file

        Remark:
        - MODBUS/TCP Security Protocol Specification demands TLSv2 at least
        - verify_mode is set to ssl.NONE
        """
        return CommParams.generate_ssl(
            False, certfile=certfile, keyfile=keyfile, password=password
        )

class ModbusTlsClient(ModbusTcpClient):
    """**ModbusTlsClient**.

    Fixed parameters:

    :param host: Host IP address or host name

    Optional parameters:

    :param sslctx: SSLContext to use for TLS
    :param framer: Framer name, default FramerType.TLS
    :param port: Port used for communication
    :param name: Set communication name, used in logging
    :param source_address: Source address of client
    :param reconnect_delay: Not used in the sync client
    :param reconnect_delay_max: Not used in the sync client
    :param timeout: Timeout for connecting and receiving data, in seconds.
    :param retries: Max number of retries per request.
    :param trace_packet: Called with bytestream received/to be sent
    :param trace_pdu: Called with PDU received/to be sent
    :param trace_connect: Called when connected/disconnected

    .. tip::
        The trace methods allow to modify the datastream/pdu !

    Example::

        from pymodbus.client import ModbusTlsClient

        async def run():
            client = ModbusTlsClient("localhost")

            client.connect()
            ...
            client.close()

    Please refer to :ref:`Pymodbus internals` for advanced usage.
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        host: str,
        *,
        sslctx: ssl.SSLContext = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT),
        framer: FramerType = FramerType.TLS,
        port: int = 802,
        name: str = "comm",
        source_address: tuple[str, int] | None = None,
        reconnect_delay: float = 0.1,
        reconnect_delay_max: float = 300,
        timeout: float = 3,
        retries: int = 3,
        trace_packet: Callable[[bool, bytes], bytes] | None = None,
        trace_pdu: Callable[[bool, ModbusPDU], ModbusPDU] | None = None,
        trace_connect: Callable[[bool], None] | None = None,
    ):
        """Initialize Modbus TLS Client."""
        if framer not in [FramerType.TLS]:
            raise TypeError("Only FramerType TLS allowed.")
        self.comm_params = CommParams(
            comm_type=CommType.TLS,
            host=host,
            sslctx=sslctx,
            port=port,
            comm_name=name,
            source_address=source_address,
            reconnect_delay=reconnect_delay,
            reconnect_delay_max=reconnect_delay_max,
            timeout_connect=timeout,
        )
        super().__init__(
            "",
            framer=framer,
            retries=retries,
            trace_packet=trace_packet,
            trace_pdu=trace_pdu,
            trace_connect=trace_connect,
        )

    @classmethod
    def generate_ssl(
        cls,
        certfile: str | None = None,
        keyfile: str | None = None,
        password: str | None = None,
    ) -> ssl.SSLContext:
        """Generate sslctx from cert/key/password.

        :param certfile: Cert file path for TLS server request
        :param keyfile: Key file path for TLS server request
        :param password: Password for for decrypting private key file

        Remark:
        - MODBUS/TCP Security Protocol Specification demands TLSv2 at least
        - verify_mode is set to ssl.NONE
        """
        return CommParams.generate_ssl(
            False, certfile=certfile, keyfile=keyfile, password=password,
        )

    @property
    def connected(self) -> bool:
        """Connect internal."""
        return self.socket is not None

    def connect(self):
        """Connect to the modbus tls server."""
        if self.socket:
            return True
        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            if self.comm_params.source_address:
                sock.bind(self.comm_params.source_address)
            self.socket = self.comm_params.sslctx.wrap_socket(sock, server_side=False)  # type: ignore[union-attr]
            self.socket.settimeout(self.comm_params.timeout_connect)
            self.socket.connect((self.comm_params.host, self.comm_params.port))
        except OSError as msg:
            Log.error(
                "Connection to ({}, {}) failed: {}",
                self.comm_params.host,
                self.comm_params.port,
                msg,
            )
            self.close()
        return self.socket is not None

    def __repr__(self):
        """Return string representation."""
        return (
            f"<{self.__class__.__name__} at {hex(id(self))} socket={self.socket}, "
            f"ipaddr={self.comm_params.host}, port={self.comm_params.port}, sslctx={self.comm_params.sslctx}, "
            f"timeout={self.comm_params.timeout_connect}>"
        )



# -*- coding: utf-8 -*-

"""package scinstr.daq.labjack
author    Benoit Dubois
copyright FEMTO Engineering, 2020-2024
licence   GPL 3.0+
brief     Handle Labjack T7(-Pro) device (DAQ board) through ethernet
          interface.
"""

import logging
import pymodbus.exceptions as mdex
from pymodbus.client.sync import ModbusTcpClient

from scinstr.daq.labjack.t7 import T7


# =============================================================================
class T7Eth(T7):
    """T7Eth class, provide command/response handling of LabJack T7-Pro
    board through ethernet interface.
    """

    def __init__(self, ip=""):
        """Constructor.
        :param ip: IP address of T7-Pro device (str)
        :returns: None
        """
        super().__init__()
        self.ip = ip

    def connect(self):
        """Connect to device.
        :returns: None
        """
        logging.info('Connecting to T7 @%s...', self.ip)
        self._client = ModbusTcpClient(self.ip)
        retval = self._client.connect()
        if retval is True:
            logging.info('Connection --> Ok')
            return True
        logging.info('Connection --> Nok')
        return False

    def close(self):
        """Close connection with device.
        :returns: None
        """
        try:
            self._client.close()
        except mdex.ModbusException as ex:
            logging.error("Modbus error %r", ex)
        except Exception as ex:
            logging.error("Unexpected error %r", ex)
        logging.info("Connection to T7 closed")


# =============================================================================
if __name__ == "__main__":
    # For "Ctrl+C" works
    import signal
    signal.signal(signal.SIGINT, signal.SIG_DFL)

    # Setup logger
    LOG_FORMAT = '%(asctime)s %(levelname)s %(filename)s (%(lineno)d): ' \
                 + '%(message)s'
    logging.basicConfig(format=LOG_FORMAT, level=logging.INFO)

    import time
    from os.path import expanduser

    I0 = 200e-6
    IP = '192.168.0.206'
    HOME = expanduser("~")
    FILE = HOME + "/t7_test.dat"

    def pt100_v2t(volt, i0):
        """Temperature from voltage of a Pt100 sensor.
        """
        return (2.604 * volt / i0) - 260.4

    T7 = T7Eth(IP)
    T7.connect()
    print("Id:", T7.get_id())

    AINS = [0, 1, 2, 3]
    T7.set_ains_resolution(AINS, [8, 8, 8, 8])

    with open(FILE, "a") as fd:
        while True:
            V_RAW = T7.get_ains_voltage(AINS)
            V_STR = ''
            for V in V_RAW:
                V_STR += str(V) + ';'
            print(V_STR)
            time.sleep(1.0)

    T7.close()
    exit()



"""Modbus Socket frame implementation."""
from __future__ import annotations

from pymodbus.framer.base import FramerBase
from pymodbus.logging import Log


class FramerSocket(FramerBase):
    """Modbus Socket frame type.

    Layout::

        [         MBAP Header         ] [ Function Code] [ Data ]
        [ tid ][ pid ][ length ][ uid ]
          2b     2b     2b        1b           1b           Nb

    length = uid + function code + data
    """

    MIN_SIZE = 8

    def decode(self, data: bytes) -> tuple[int, int, int, bytes]:
        """Decode ADU."""
        if (data_len := len(data)) < self.MIN_SIZE:
          Log.debug("Very short frame (NO MBAP): {} wait for more data", data, ":hex")
          return 0, 0, 0, self.EMPTY
        tid = int.from_bytes(data[0:2], 'big')
        msg_len = int.from_bytes(data[4:6], 'big') + 6
        dev_id = int(data[6])
        if data_len < msg_len:
          Log.debug("Short frame: {} wait for more data", data, ":hex")
          return 0, 0, 0, self.EMPTY
        if msg_len == 8 and data_len == 9:
            msg_len = 9
        return msg_len, dev_id, tid, data[7:msg_len]

    def encode(self, pdu: bytes, device_id: int, tid: int) -> bytes:
        """Encode ADU."""
        frame = (
           tid.to_bytes(2, 'big') +
           b'\x00\x00' +
           (len(pdu) + 1).to_bytes(2, 'big') +
           device_id.to_bytes(1, 'big') +
           pdu
        )
        return frame



# Set __version__ in the setup.py
with open('scinstr/version.py') as f: exec(f.read())

from setuptools import setup

setup(name='scinstr',
      description='Drivers to handle various scientific instruments (DMM, frequency counter, DAQ, VNA...). Include also some cli script.',
      version=__version__,
      packages=['scinstr.cnt',
                'scinstr.compressor',
                'scinstr.daq',
                'scinstr.daq.labjack',
                'scinstr.dmm',
                'scinstr.lockin',
                'scinstr.phimeter',
                'scinstr.tctrl',
                'scinstr.vacuum',
                'scinstr.vna'],
      install_requires=['pyusb',
                        'PyVISA',
                        'PyVISA-py',
                        'pyserial',
                        'python-usbtmc',
                        'pymodbus'],
      extras_require={
          'Pure_Python_signalslot_facilities': ["signalslot"],
      },
      url='https://gitlab.com/bendub/scinstr',
      author='Benoit Dubois',
      author_email='benoit.dubois@femto-engineering.fr',
      classifiers=[
          'Development Status :: 4 - Beta',
          'Environment :: Console',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)',
          'Natural Language :: English',
          'Operating System :: POSIX :: Linux',
          'Programming Language :: Python',
          'Topic :: Scientific/Engineering'
          ]
)



"""Remote datastore."""
from pymodbus.datastore import ModbusBaseSlaveContext
from pymodbus.exceptions import NotImplementedException
from pymodbus.logging import Log


# ---------------------------------------------------------------------------#
#  Context
# ---------------------------------------------------------------------------#
class RemoteSlaveContext(ModbusBaseSlaveContext):
    """TODO.

    This creates a modbus data model that connects to
    a remote device (depending on the client used)
    """

    def __init__(self, client, slave=None):
        """Initialize the datastores.

        :param client: The client to retrieve values with
        :param slave: Unit ID of the remote slave
        """
        self._client = client
        self.slave = slave
        self.result = None
        self.__build_mapping()
        if not self.__set_callbacks:
            Log.error("Init went wrong.")

    def reset(self):
        """Reset all the datastores to their default values."""
        raise NotImplementedException()

    def getValues(self, fc_as_hex, _address, _count=1):
        """Get values from remote device."""
        if fc_as_hex in self._write_fc:
            return [0]
        group_fx = self.decode(fc_as_hex)
        func_fc = self.__get_callbacks[group_fx]
        self.result = func_fc(_address, _count)
        return self.__extract_result(self.decode(fc_as_hex), self.result)

    def setValues(self, fc_as_hex, address, values):
        """Set the datastore with the supplied values."""
        group_fx = self.decode(fc_as_hex)
        if fc_as_hex not in self._write_fc:
            raise ValueError(f"setValues() called with an non-write function code {fc_as_hex}")
        func_fc = self.__set_callbacks[f"{group_fx}{fc_as_hex}"]
        if fc_as_hex in {0x0F, 0x10}:  # Write Multiple Coils, Write Multiple Registers
            self.result = func_fc(address, values)
        else:
            self.result = func_fc(address, values[0])
        # if self.result.isError():
        #    return self.result

    def __str__(self):
        """Return a string representation of the context.

        :returns: A string representation of the context
        """
        return f"Remote Slave Context({self._client})"

    def __build_mapping(self):
        """Build the function code mapper."""
        params = {}
        if self.slave:
            params["slave"] = self.slave
        self.__get_callbacks = {
            "d": lambda a, c: self._client.read_discrete_inputs(
                a, count=c, **params
            ),
            "c": lambda a, c: self._client.read_coils(
                a, count=c, **params
            ),
            "h": lambda a, c: self._client.read_holding_registers(
                a, count=c, **params
            ),
            "i": lambda a, c: self._client.read_input_registers(
                a, count=c, **params
            ),
        }
        self.__set_callbacks = {
            "d5": lambda a, v: self._client.write_coil(
                a, v, **params
            ),
            "d15": lambda a, v: self._client.write_coils(
                a, v, **params
            ),
            "c5": lambda a, v: self._client.write_coil(
                a, v, **params
            ),
            "c15": lambda a, v: self._client.write_coils(
                a, v, **params
            ),
            "h6": lambda a, v: self._client.write_register(
                a, v, **params
            ),
            "h16": lambda a, v: self._client.write_registers(
                a, v, **params
            ),
            "i6": lambda a, v: self._client.write_register(
                a, v, **params
            ),
            "i16": lambda a, v: self._client.write_registers(
                a, v, **params
            ),
        }
        self._write_fc = (0x05, 0x06, 0x0F, 0x10)

    def __extract_result(self, fc_as_hex, result):
        """Extract the values out of a response.

        TODO make this consistent (values?)
        """
        if not result.isError():
            if fc_as_hex in {"d", "c"}:
                return result.bits
            if fc_as_hex in {"h", "i"}:
                return result.registers
        else:
            return result
        return None



"""Transport."""
__all__ = [
    "NULLMODEM_HOST",
    "CommParams",
    "CommType",
    "ModbusProtocol",
]

from pymodbus.transport.transport import (
    NULLMODEM_HOST,
    CommParams,
    CommType,
    ModbusProtocol,
)



"""Diagnostic record read/write."""
from __future__ import annotations

import struct

from pymodbus.constants import ModbusStatus
from pymodbus.datastore import ModbusSlaveContext
from pymodbus.device import DeviceInformationFactory, ModbusControlBlock
from pymodbus.pdu.pdu import ModbusPDU


_MCB = ModbusControlBlock()


class ReadExceptionStatusRequest(ModbusPDU):
    """ReadExceptionStatusRequest."""

    function_code = 0x07
    rtu_frame_size = 4

    def encode(self) -> bytes:
        """Encode the message."""
        return b""

    def decode(self, _data: bytes) -> None:
        """Decode data part of the message."""

    async def update_datastore(self, _context: ModbusSlaveContext) -> ModbusPDU:
        """Run a read exception status request against the store."""
        status = _MCB.Counter.summary()
        return ReadExceptionStatusResponse(status=status, dev_id=self.dev_id, transaction_id=self.transaction_id)


class ReadExceptionStatusResponse(ModbusPDU):
    """ReadExceptionStatusResponse."""

    function_code = 0x07
    rtu_frame_size = 5

    def encode(self) -> bytes:
        """Encode the response."""
        return struct.pack(">B", self.status)

    def decode(self, data: bytes) -> None:
        """Decode a the response."""
        self.status = int(data[0])


# Encapsulate interface transport 43, 14
# CANopen general reference 43, 13

class GetCommEventCounterRequest(ModbusPDU):
    """GetCommEventCounterRequest."""

    function_code = 0x0B
    rtu_frame_size = 4

    def encode(self) -> bytes:
        """Encode the message."""
        return b""

    def decode(self, _data: bytes) -> None:
        """Decode data part of the message."""

    async def update_datastore(self, _context) -> ModbusPDU:
        """Run a read exception status request against the store."""
        count = _MCB.Counter.Event
        return GetCommEventCounterResponse(count=count, dev_id=self.dev_id, transaction_id=self.transaction_id)


class GetCommEventCounterResponse(ModbusPDU):
    """GetCommEventCounterRequest."""

    function_code = 0x0B
    rtu_frame_size = 8

    def encode(self) -> bytes:
        """Encode the response."""
        ready = ModbusStatus.READY if self.status else ModbusStatus.WAITING
        return struct.pack(">HH", ready, self.count)

    def decode(self, data: bytes) -> None:
        """Decode a the response."""
        ready, self.count = struct.unpack(">HH", data)
        self.status = ready == ModbusStatus.READY


class GetCommEventLogRequest(ModbusPDU):
    """GetCommEventLogRequest."""

    function_code = 0x0C
    rtu_frame_size = 4

    def encode(self) -> bytes:
        """Encode the message."""
        return b""

    def decode(self, _data: bytes) -> None:
        """Decode data part of the message."""

    async def update_datastore(self, _context: ModbusSlaveContext) -> ModbusPDU:
        """Run a read exception status request against the store."""
        return GetCommEventLogResponse(
            status=True,
            message_count=_MCB.Counter.BusMessage,
            event_count=_MCB.Counter.Event,
            events=_MCB.getEvents(),
            dev_id=self.dev_id, transaction_id=self.transaction_id)


class GetCommEventLogResponse(ModbusPDU):
    """GetCommEventLogRequest."""

    function_code = 0x0C
    rtu_byte_count_pos = 2

    def __init__(self, status: bool = True, message_count: int = 0, event_count: int = 0, events: list[int] | None = None, dev_id: int = 1, transaction_id: int = 0) -> None:
        """Initialize a new instance."""
        super().__init__(transaction_id=transaction_id, dev_id=dev_id, status=status)
        self.message_count = message_count
        self.event_count = event_count
        self.events = events or []

    def encode(self) -> bytes:
        """Encode the response."""
        if self.status:
            ready = ModbusStatus.READY
        else:
            ready = ModbusStatus.WAITING
        packet = struct.pack(">B", 6 + len(self.events))
        packet += struct.pack(">H", ready)
        packet += struct.pack(">HH", self.event_count, self.message_count)
        packet += b"".join(struct.pack(">B", e) for e in self.events)
        return packet

    def decode(self, data: bytes) -> None:
        """Decode a the response."""
        length = int(data[0])
        status = struct.unpack(">H", data[1:3])[0]
        self.status = status == ModbusStatus.READY
        self.event_count = struct.unpack(">H", data[3:5])[0]
        self.message_count = struct.unpack(">H", data[5:7])[0]

        self.events = []
        for i in range(7, length + 1):
            self.events.append(int(data[i]))


class ReportSlaveIdRequest(ModbusPDU):
    """ReportSlaveIdRequest."""

    function_code = 0x11
    rtu_frame_size = 4

    def encode(self) -> bytes:
        """Encode the message."""
        return b""

    def decode(self, _data: bytes) -> None:
        """Decode data part of the message."""

    async def update_datastore(self, _context: ModbusSlaveContext) -> ModbusPDU:
        """Run a report slave id request against the store."""
        information = DeviceInformationFactory.get(_MCB)
        id_data = []
        for v_item in information.values():
            if isinstance(v_item, bytes):
                id_data.append(v_item)
            else:
                id_data.append(v_item.encode())

        identifier = b"-".join(id_data)
        identifier = identifier or b"Pymodbus"
        return ReportSlaveIdResponse(identifier=identifier, dev_id=self.dev_id, transaction_id=self.transaction_id)


class ReportSlaveIdResponse(ModbusPDU):
    """ReportSlaveIdRequeste."""

    function_code = 0x11
    rtu_byte_count_pos = 2

    def __init__(self, identifier: bytes = b"\x00", status: bool = True, dev_id: int = 1, transaction_id: int = 0) -> None:
        """Initialize a new instance."""
        super().__init__(transaction_id=transaction_id, dev_id=dev_id, status=status)
        self.identifier = identifier
        self.byte_count = 0

    def encode(self) -> bytes:
        """Encode the response."""
        status = ModbusStatus.SLAVE_ON if self.status else ModbusStatus.SLAVE_OFF
        length = len(self.identifier) + 1
        packet = struct.pack(">B", length)
        packet += self.identifier  # we assume it is already encoded
        packet += struct.pack(">B", status)
        return packet

    def decode(self, data: bytes) -> None:
        """Decode a the response.

        Since the identifier is device dependent, we just return the
        raw value that a user can decode to whatever it should be.
        """
        self.byte_count = int(data[0])
        self.identifier = data[1 : self.byte_count + 1]
        status = int(data[-1])
        self.status = status == ModbusStatus.SLAVE_ON



# -*- coding: utf-8 -*-

"""package scinstr
author    Benoit Dubois
copyright FEMTO ENGINEERING, 2019-2024
license   GPL v3.0+
brief     API dedicated to handle the Keysight 34461A digital multimeter.
"""

import logging
import time
import struct
import usb
import pyvisa

from scinstr.dmm.constants import (
    SDM3065X_TIMEOUT,
    SDM3065X_ID,
    SDM3065X_VID,
    SDM3065X_PID,
)

pyvisa.ResourceManager('@py')


MAX_REQUEST_RETRIES = 5
MAX_READ_FLUSH_ATTEMPT = 10

# Multimeter functions
(
    CAP,
    CURR,
    CURRDC,
    VOLTAC,
    VOLTDC,
    VOLTDCRAT,
    RES,
    FRES,
    FRE,
    PER,
    TEMPRTD,
    TEMPFRTD,
    DIOD,
    CONT,
) = (
    "CAP",
    "CURR",
    "CURR:DC",
    "VOLT:AC",
    "VOLT:DC",
    "VOLT:DC:RAT",
    "RES",
    "FRES",
    "FRE",
    "PER",
    "TEMP:RTD",
    "TEMP:FRTD",
    "DIOD",
    "CONT",
)

# Integration time
(PLC002, PLC02, PLC1, PLC10, PLC100) = ("0.02", "0.2", "1", "10", "100")

# Voltage input range
(RANGE10MV, RANGE1V, RANGE10V, RANGE100V, RANGE1000V) = (
    "0.1",
    "1",
    "10",
    "100",
    "1000",
)

# Trigger source
(IMM, EXT, BUS) = ("IMM", "EXT", "BUS")


# =============================================================================
class Sdm3065xAbstract:
    """Abstract class to handling Sdm3065x digital multimeter device. Derived
    classes need to re-implement specific-protocol methods: connect(), close(),
    _write(), _read()...
    """

    def _write(self, data: str) -> int:
        """Abstract protocol write process. Derived classes must implement
        the write process dedicated to the specific protocol used.
        :param data: data writes to device (str)
        :returns: number of writed byte
        """
        raise NotImplementedError("Method not implemented by derived class")

    def _read(self) -> str:
        """Abstract protocol read process. Derived classes must implement
        the read process dedicated to the specific protocol used.
        :returns: Message reads from device (str)
        """
        raise NotImplementedError("Method not implemented by derived class")

    def query(self, msg) -> str:
        """Abstract protocol read process. Derived classes must implement
        the read process dedicated to the specific protocol used.
        :returns: Message reads from device (str)
        """
        raise NotImplementedError("Method not implemented by derived class")

    def write(self, data: str) -> int:
        """A basic write method: writes "data" to device.
        :param data: data writes to device (str)
        :returns: number of writed byte, -1 in case of error
        """
        try:
            retval = self._write(data)
        except Exception as ex:
            logging.error(f"Write message:{data}, error: {ex}")
            return -1
        logging.debug(f"write: {data}")
        return retval

    def read(self) -> str | None:
        """A basic read method: read a message from device.
        :param length: length of message to read (int)
        :returns: Message reads from device, None if error (str)
        """
        try:
            retval = self._read()
        except Exception as ex:
            logging.error(f"Read error: {ex}")
            return None
        logging.debug(f"read: {retval}")
        return retval

    def flush(self) -> bool:
        """Flush DMM HW buffer"""
        read_flush_attempt = MAX_READ_FLUSH_ATTEMPT
        while read_flush_attempt >= 0:
            try:
                _ = self._read()
            except usb.core.USBTimeoutError:
                return True
            read_flush_attempt -= 1

            logging.critical(f"flush({read_flush_attempt})")

        logging.error("DMM flush failed")
        return False

    def reset(self) -> bool:
        """Resets meter to its power-on state, sets all bits to zero in
        status byte register and all event registers and clear error queue.
        :returns: None
        """
        try:
            self.write("ABOR")
            self.write("*RST")
            self.write("*CLS")
        except Exception as ex:
            logging.error("reset failed %r", ex)
            return False
        time.sleep(2.0)
        logging.info("DMM reseted")
        return True

    def get_error(self) -> list[str] | None:
        """Return list of current error.
        :returns: list of current error (list of str)
        """
        errors = []
        while True:
            error = self.query("SYST:ERR?")
            if error is None or error == "":
                logging.error(f"Get error from device failed.")
                return None
            if "No error" in error:
                break
            errors.append(error)
        return errors

    def check_interface(self) -> bool:
        """Basic interface connection test: check id of device.
        Return True if interface with device is OK.
        :returns: status of interface with device (bool)
        """
        id_ = self.query("*IDN?")
        if id_ is None:
            return False
        if SDM3065X_ID in id_:
            return True
        return False

    def data_read(self) -> float | None:
        """Takes a measurement the next time the trigger condition is met.
        After the measurement is taken, the reading is placed in the output
        buffer. "data_read" will not cause readings to be stored in the Meter’s
        internal memory.
        Read method convenient for slow measurement.
        :returns: data read in buffer (float)
        """
        data = self.query("READ?")
        if data is None:
            logging.error(f"READ? returns None")
            return None
        try:
            data = float(data)
        except ValueError:
            logging.error(f"Invalid data {data}")
            return None
        return data


# =============================================================================
class Sdm3065xUsb(Sdm3065xAbstract):
    """Handle DMM device through USB connection."""

    def __init__(
        self, vendor_id=SDM3065X_VID, product_id=SDM3065X_PID, timeout=SDM3065X_TIMEOUT, serial_id=None
    ):
        rm = pyvisa.ResourceManager()
        # USB[board]::manufacturer ID::model code::serial number[::USB interface number][::INSTR]
        # USB::0x1234::125::A22-5::INSTR
        self._dev = rm.open_resource(f"USB::{vendor_id}::{product_id}::{serial_id}::INSTR", write_termination='\n', query_delay=timeout)
        self._dev.read_termination = '\n'
        self._dev.write_termination = '\n'
        self._dev.timeout = timeout * 1000

    def _write(self, data: str):
        """Specific USB writing process.
        :param data: data writes to device (str)
        :returns: number of bytes sent (int)
        """
        self._dev.write(data)
        return len(data)

    def _read(self) -> str:
        """Specific USB reading process.
        :param length: length of message to read (int)
        :returns: Message reads from device (str)
        """
        request_retry = MAX_REQUEST_RETRIES
        while True:
            try:
                msg = self._dev.read()
            except struct.error:
                print(f"request_retry: {request_retry}")
                request_retry -= 1
                if request_retry < 0:
                    raise ConnectionError("_read() failed")
                continue
            break
        return msg

    def query(self, msg):
        return self._dev.query(msg)

    def ask(self, msg):
        return self.query(msg)

    @property
    def timeout(self):
        """Get timeout on socket operations.
        :returns: timeout value in second (float)
        """
        return self._dev.timeout / 1000

    @timeout.setter
    def timeout(self, timeout):
        """Set timeout on socket operations.
        :param timeout: timeout value in second (float)
        :returns: None
        """
        self._dev.timeout = timeout * 1000


# =============================================================================
def check_dmm():
    """Check the Dmm34461axx class: connect to the multimeter, configure a dc
    voltage measurement then collect and print data to standard output.
    """
    import datetime

    date_fmt = "%d/%m/%Y %H:%M:%S"
    log_format = "%(asctime)s %(levelname) -8s %(filename)s " + \
                 " %(funcName)s (%(lineno)d): %(message)s"
    logging.basicConfig(level=logging.INFO,
                        datefmt=date_fmt,
                        format=log_format)

    from scinstr.dmm.constants import SDM3055_VID as VID
    from scinstr.dmm.constants import SDM3055_PID as PID

    SERIAL_ID = "SDM35HBQ802983"

    # rm = pyvisa.ResourceManager()
    # print(f"list_devices(): {rm.list_resources()}")

    dmm = Sdm3065xUsb(VID, PID, 2.1, SERIAL_ID)
    print("IDN:", dmm.query("*IDN?"))

    dmm.write("CONF:VOLT:DC")
    dmm.write("CONF:VOLT:DC auto")  # Range
    dmm.write("VOLT:DC:NPLC 10")
    dmm.write("VOLT:DC:AZ:STATE ON")  # Autozero state
    dmm.write("SAMP:COUN 1")
    dmm.write("TRIG:COUN 10")
    dmm.write("TRIG:SOUR IMM")
    dmm.write("INIT")
    # print("Error config?:", dmm.get_error())

    try:
        while True:
            try:
                values = dmm.query("READ?")
                value = sum([float(x) for x in values.split(',')]) / len(values)
                # value = dmm.query("MEAS:VOLT:DC? auto")
                now = datetime.datetime.now(datetime.UTC)
                if value is None or value == "":
                    print("# No data @", now)
                else:
                    print(now, value)
            except KeyboardInterrupt:
                break
            except Exception as er:
                logging.error("# Exception during acquisition: %r", er)
    except KeyboardInterrupt:
        dmm.write("ABORT")

    print("Final error?:", dmm.get_error())


# =============================================================================
if __name__ == '__main__':
    check_dmm()


# -*- coding: utf-8 -*-

"""package scinstr
author    Benoit Dubois
copyright FEMTO ENGINEERING, 2019-2025
license   GPL v3.0+
brief     Emulate basic SCPI counter with state tuningable from webserver
"""

import logging
import threading
import time
import random
from enum import Enum
from remi import start, App, gui

# Only to mime real device import (detection of circular reference)
import socket
import usbtmc


DEFAULT_PORT = 8091
DEFAULT_FREQUENCY = 99999999.925  #100E6  # + 1.0


CommunicationState = Enum('State', [("OPENED", 1), ("CLOSED", 2), ("FAULTED", 3)])


# =============================================================================
def mystart(class_, port, userdata):
    start(class_, port=port, userdata=userdata)


# =============================================================================
class CntEmulator:
    """Emulate a basic Counter device behavior and problem:
    - get (emulated) measured values (user defined or randomized)
    - emulate disconnection/reconnection with device
    """

    def __init__(self, *args, **kwargs):
        """The constructor.
        :returns: None
        """
        self._log = logging.getLogger('scinstr.cnt.emul')
        self._communication_state = CommunicationState.CLOSED
        self._is_random_data = False
        if 'f_mean' in kwargs.keys():
            self._freq = kwargs['f_mean']
        else:
            self._freq = DEFAULT_FREQUENCY
        for idx, value in enumerate(args):
            self._log.info('Counter test device non-keyworded argument %02d: %r',
                         idx, value)
        for key, value in kwargs.items():
            self._log.info('Counter test device named argument %r: %r', key, value)
        mythread = threading.Thread(target=mystart, args=(CntEmulatorHandler, DEFAULT_PORT, (self,)))
        mythread.start()
        self._log.info("Counter test device %r initialization done", self)

    def set_frequency(self, value: float) -> None:
        self._freq = value
        self._log.info(f"set_frequency {value}")

    def set_communication_state(self, value: CommunicationState)-> None:
        self._communication_state = value
        self._log.info(f"set_communication_state {value}")

    def set_random_data_state(self, state)-> None:
        self._is_random_data = state

    @property
    def timeout(self):
        """Gets timeout on socket operations.
        :returns: timeout value in second (float)
        """
        self._log.info("Get Counter test device timeout: %r", self._timeout)
        return self._timeout

    @timeout.setter
    def timeout(self, timeout):
        """Sets timeout on socket operations.
        :param timeout: timeout value in second (float)
        :returns: None
        """
        self._timeout = timeout
        self._log.info("Set Counter test device timeout: %r", timeout)

    def connect(self) -> bool:
        """Connection process to Counter.
        :returns: True if connection success other False (Bool)
        """
        self._communication_state = CommunicationState.OPENED
        self._log.info("Connected to Counter test device: %r", self)
        return True

    def close(self)-> None:
        """Closing process with DMM.
        :returns: None
        """
        self._communication_state = CommunicationState.CLOSED
        self._log.info("Connection to DMM emulator device %r closed", self)

    def read(self, length):
        """Emulate read process.
        :param length: length of message to read (int)
        :returns: Message reads from device (str)
        """
        if self._communication_state == CommunicationState.FAULTED:
            raise ConnectionError("Communication with (emulated) DMM failled")
        data = "emulated_counter"
        self._log.info("Read %r from DMM emulator device %r", data, self)
        return data

    def write(self, data):
        """Emulate write process
        :param data: data writes to device (str)
        :returns: None
        """
        if self._communication_state == CommunicationState.FAULTED:
            raise ConnectionError("Communication with (emulated) DMM failled")
        self._log.debug("Write %r to Counter test device %r", data, self)

    def query(self, msg):
        self._log.info(f"query({msg})")
        if msg == "SYST:ERR?":
            return "No error"
        return msg

    def reset(self) -> bool:
        """Connection process to Counter.
        :returns: True if connection success other False (Bool)
        """
        self._log.info("Reset Counter test device: %r", self)
        return True

    def get_error(self)-> list[str]:
        """Subclass method to emulate response of device.
        """
        return ["+0,\"No error\""]

    @property
    def is_connected(self):
        match self._communication_state:
            case CommunicationState.OPENED:
                return True
            case CommunicationState.CLOSED:
                return False
            case _:  # i.e. CommunicationState.FAULTED
                raise ConnectionError("Connection with Counter emulator failed")

    def data_read(self):
        if self._communication_state == CommunicationState.FAULTED:
            raise ConnectionError("Communication with (emulated) Counter failled")
        time.sleep(1.0)  # Acquisition time
        data = self._freq
        if self._is_random_data is True:
            data += self._random_gen()
        self._log.debug("Counter test data_read(%r) %r", data, self)
        return data

    def _random_gen(self) -> float:
        """Return a random value: use to emulate data input acquisition.
        """
        return random.uniform(-1.0, 1.0)


# =============================================================================
class CntEmulatorHandler(App):
    """Handle emulated Counter state from a web app.
    """

    def main(self, counter)-> gui.Widget:
        self._cnt = counter
        #
        main_box = gui.VBox()
        title_box = gui.HBox()
        title_lbl = gui.Label('Counter emulator handler', width="80%", margin="5px")
        comm_fault_box = gui.HBox()
        comm_fault_lbl = gui.Label('Emulate communication fault', width="80%", margin="5px")
        self.comm_fault_cbox = gui.CheckBox(False, width="20%", margin="5px")
        freq_box = gui.HBox()
        freq_lbl = gui.Label('Frequency value (Hz)', width="20%", margin="5px")
        self.freq_input = gui.TextInput(width="40%", margin="5px")
        self.freq_input.set_text(str(DEFAULT_FREQUENCY))
        self.freq_btn = gui.Button('Set', width="10%", margin="5px")
        #
        title_box.append(title_lbl)
        comm_fault_box.append(comm_fault_lbl)
        comm_fault_box.append(self.comm_fault_cbox)
        freq_box.append(freq_lbl)
        freq_box.append(self.freq_input)
        freq_box.append(self.freq_btn)
        main_box.append(title_box)
        main_box.append(comm_fault_box)
        main_box.append(freq_box)
        #
        self.comm_fault_cbox.onclick.do(self.comm_fault_pressed)
        self.freq_btn.onclick.do(self.freq_btn_pressed)
        #
        return main_box

    def comm_fault_pressed(self, widget)-> None:
        if self.comm_fault_cbox.get_value() is True:
            state = CommunicationState.FAULTED
        else:
            state = CommunicationState.OPENED
        self._cnt.set_communication_state(state)

    def freq_btn_pressed(self, widget)-> None:
        self._log.critical(f"set frequency: {self.freq_input.text}")
        self._cnt.set_frequency(self.freq_input.text)


# =============================================================================
def check_cnt_emulator():
    """Check the Cnt532x0aEmul class: connect to the counter, configure
    frequency measurement then collect and print data to standard output.
    """
    from datetime import datetime

    date_fmt = "%d/%m/%Y %H:%M:%S"
    log_format = "%(asctime)s %(levelname) -8s %(filename)s " + \
                 " %(funcName)s (%(lineno)d): %(message)s"
    self._log.basicConfig(level=self._log.INFO,
                        datefmt=date_fmt,
                        format=log_format)

    f_mean = 10E7

    cnt = CntEmulator(0x2a8d, 0x1601, timeout=4.8, f_mean=f_mean)



if __name__ == '__main__':
    check_cnt_emulator()



"""Framer implementations.

The implementation is responsible for encoding/decoding requests/responses.

According to the selected type of modbus frame a prefix/suffix is added/removed
"""
from __future__ import annotations

from enum import Enum

from pymodbus.exceptions import ModbusIOException
from pymodbus.logging import Log
from pymodbus.pdu import DecodePDU, ModbusPDU


class FramerType(str, Enum):
    """Type of Modbus frame."""

    ASCII = "ascii"
    RTU = "rtu"
    SOCKET = "socket"
    TLS = "tls"


class FramerBase:
    """Intern base."""

    EMPTY = b''
    MIN_SIZE = 0

    def __init__(
        self,
        decoder: DecodePDU,
    ) -> None:
        """Initialize a ADU (framer) instance."""
        self.decoder = decoder

    def decode(self, _data: bytes) -> tuple[int, int, int, bytes]:
        """Decode ADU.

        returns:
            used_len (int) or 0 to read more
            dev_id,
            tid,
            modbus request/response (bytes)
        """
        return 0, 0, 0, self.EMPTY

    def encode(self, data: bytes, _dev_id: int, _tid: int) -> bytes:
        """Encode ADU.

        returns:
            modbus ADU (bytes)
        """
        return data

    def buildFrame(self, message: ModbusPDU) -> bytes:
        """Create a ready to send modbus packet.

        :param message: The populated request/response to send
        """
        data = message.function_code.to_bytes(1,'big') + message.encode()
        frame = self.encode(data, message.dev_id, message.transaction_id)
        return frame

    def processIncomingFrame(self, data: bytes) -> tuple[int, ModbusPDU | None]:
        """Process new packet pattern.

        This takes in a new request packet, adds it to the current
        packet stream, and performs framing on it. That is, checks
        for complete messages, and once found, will process all that
        exist.
        """
        used_len = 0
        while True:
            data_len, pdu = self._processIncomingFrame(data[used_len:])
            used_len += data_len
            if not data_len:
                return used_len, None
            if pdu:
                return used_len, pdu

    def _processIncomingFrame(self, data: bytes) -> tuple[int, ModbusPDU | None]:
        """Process new packet pattern.

        This takes in a new request packet, adds it to the current
        packet stream, and performs framing on it. That is, checks
        for complete messages, and once found, will process all that
        exist.
        """
        Log.debug("Processing: {}", data, ":hex")
        if not data:
            return 0, None
        used_len, dev_id, tid, frame_data = self.decode(data)
        if not frame_data:
            return used_len, None
        if (result := self.decoder.decode(frame_data)) is None:
            raise ModbusIOException("Unable to decode request")
        result.dev_id = dev_id
        result.transaction_id = tid
        Log.debug("Frame advanced, resetting header!!")
        return used_len, result



import sys
from optparse import Values
from typing import AbstractSet, List

from pip._internal.cli import cmdoptions
from pip._internal.cli.base_command import Command
from pip._internal.cli.status_codes import SUCCESS
from pip._internal.operations.freeze import freeze
from pip._internal.utils.compat import stdlib_pkgs


def _should_suppress_build_backends() -> bool:
    return sys.version_info < (3, 12)


def _dev_pkgs() -> AbstractSet[str]:
    pkgs = {"pip"}

    if _should_suppress_build_backends():
        pkgs |= {"setuptools", "distribute", "wheel"}
        pkgs |= {"setuptools", "distribute", "wheel", "pkg-resources"}

    return pkgs


class FreezeCommand(Command):
    """
    Output installed packages in requirements format.

    packages are listed in a case-insensitive sorted order.
    """

    ignore_require_venv = True
    usage = """
      %prog [options]"""
    log_streams = ("ext://sys.stderr", "ext://sys.stderr")

    def add_options(self) -> None:
        self.cmd_opts.add_option(
            "-r",
            "--requirement",
            dest="requirements",
            action="append",
            default=[],
            metavar="file",
            help=(
                "Use the order in the given requirements file and its "
                "comments when generating output. This option can be "
                "used multiple times."
            ),
        )
        self.cmd_opts.add_option(
            "-l",
            "--local",
            dest="local",
            action="store_true",
            default=False,
            help=(
                "If in a virtualenv that has global access, do not output "
                "globally-installed packages."
            ),
        )
        self.cmd_opts.add_option(
            "--user",
            dest="user",
            action="store_true",
            default=False,
            help="Only output packages installed in user-site.",
        )
        self.cmd_opts.add_option(cmdoptions.list_path())
        self.cmd_opts.add_option(
            "--all",
            dest="freeze_all",
            action="store_true",
            help=(
                "Do not skip these packages in the output:"
                " {}".format(", ".join(_dev_pkgs()))
            ),
        )
        self.cmd_opts.add_option(
            "--exclude-editable",
            dest="exclude_editable",
            action="store_true",
            help="Exclude editable package from output.",
        )
        self.cmd_opts.add_option(cmdoptions.list_exclude())

        self.parser.insert_option_group(0, self.cmd_opts)

    def run(self, options: Values, args: List[str]) -> int:
        skip = set(stdlib_pkgs)
        if not options.freeze_all:
            skip.update(_dev_pkgs())

        if options.excludes:
            skip.update(options.excludes)

        cmdoptions.check_list_path_option(options)

        for line in freeze(
            requirement=options.requirements,
            local_only=options.local,
            user_only=options.user,
            paths=options.path,
            isolated=options.isolated_mode,
            skip=skip,
            exclude_editable=options.exclude_editable,
        ):
            sys.stdout.write(line + "\n")
        return SUCCESS



"""Framer."""
__all__ = [
    "FramerAscii",
    "FramerBase",
    "FramerRTU",
    "FramerSocket",
    "FramerTLS",
    "FramerType"
]

from pymodbus.framer.ascii import FramerAscii
from pymodbus.framer.base import FramerBase, FramerType
from pymodbus.framer.rtu import FramerRTU
from pymodbus.framer.socket import FramerSocket
from pymodbus.framer.tls import FramerTLS


FRAMER_NAME_TO_CLASS = {
    FramerType.ASCII: FramerAscii,
    FramerType.RTU: FramerRTU,
    FramerType.SOCKET: FramerSocket,
    FramerType.TLS: FramerTLS,
}



"""Modbus Request/Response Decoders."""
from __future__ import annotations

import pymodbus.pdu.bit_message as bit_msg
import pymodbus.pdu.diag_message as diag_msg
import pymodbus.pdu.file_message as file_msg
import pymodbus.pdu.mei_message as mei_msg
import pymodbus.pdu.other_message as o_msg
import pymodbus.pdu.pdu as base
import pymodbus.pdu.register_message as reg_msg
from pymodbus.exceptions import MessageRegisterException, ModbusException
from pymodbus.logging import Log


class DecodePDU:
    """Decode pdu requests/responses (server/client)."""

    _pdu_class_table: set[tuple[type[base.ModbusPDU], type[base.ModbusPDU]]] = {
        (reg_msg.ReadHoldingRegistersRequest, reg_msg.ReadHoldingRegistersResponse),
        (bit_msg.ReadDiscreteInputsRequest, bit_msg.ReadDiscreteInputsResponse),
        (reg_msg.ReadInputRegistersRequest, reg_msg.ReadInputRegistersResponse),
        (bit_msg.ReadCoilsRequest, bit_msg.ReadCoilsResponse),
        (bit_msg.WriteMultipleCoilsRequest, bit_msg.WriteMultipleCoilsResponse),
        (reg_msg.WriteMultipleRegistersRequest, reg_msg.WriteMultipleRegistersResponse),
        (reg_msg.WriteSingleRegisterRequest, reg_msg.WriteSingleRegisterResponse),
        (bit_msg.WriteSingleCoilRequest, bit_msg.WriteSingleCoilResponse),
        (reg_msg.ReadWriteMultipleRegistersRequest, reg_msg.ReadWriteMultipleRegistersResponse),
        (diag_msg.DiagnosticBase, diag_msg.DiagnosticBase),
        (o_msg.ReadExceptionStatusRequest, o_msg.ReadExceptionStatusResponse),
        (o_msg.GetCommEventCounterRequest, o_msg.GetCommEventCounterResponse),
        (o_msg.GetCommEventLogRequest, o_msg.GetCommEventLogResponse),
        (o_msg.ReportSlaveIdRequest, o_msg.ReportSlaveIdResponse),
        (file_msg.ReadFileRecordRequest, file_msg.ReadFileRecordResponse),
        (file_msg.WriteFileRecordRequest, file_msg.WriteFileRecordResponse),
        (reg_msg.MaskWriteRegisterRequest, reg_msg.MaskWriteRegisterResponse),
        (file_msg.ReadFifoQueueRequest, file_msg.ReadFifoQueueResponse),
        (mei_msg.ReadDeviceInformationRequest, mei_msg.ReadDeviceInformationResponse),
    }

    _pdu_sub_class_table: set[tuple[type[base.ModbusPDU], type[base.ModbusPDU]]] = {
        (diag_msg.ReturnQueryDataRequest, diag_msg.ReturnQueryDataResponse),
        (diag_msg.RestartCommunicationsOptionRequest, diag_msg.RestartCommunicationsOptionResponse),
        (diag_msg.ReturnDiagnosticRegisterRequest, diag_msg.ReturnDiagnosticRegisterResponse),
        (diag_msg.ChangeAsciiInputDelimiterRequest, diag_msg.ChangeAsciiInputDelimiterResponse),
        (diag_msg.ForceListenOnlyModeRequest, diag_msg.ForceListenOnlyModeResponse),
        (diag_msg.ClearCountersRequest, diag_msg.ClearCountersResponse),
        (diag_msg.ReturnBusMessageCountRequest, diag_msg.ReturnBusMessageCountResponse),
        (diag_msg.ReturnBusCommunicationErrorCountRequest, diag_msg.ReturnBusCommunicationErrorCountResponse),
        (diag_msg.ReturnBusExceptionErrorCountRequest, diag_msg.ReturnBusExceptionErrorCountResponse),
        (diag_msg.ReturnSlaveMessageCountRequest, diag_msg.ReturnSlaveMessageCountResponse),
        (diag_msg.ReturnSlaveNoResponseCountRequest, diag_msg.ReturnSlaveNoResponseCountResponse),
        (diag_msg.ReturnSlaveNAKCountRequest, diag_msg.ReturnSlaveNAKCountResponse),
        (diag_msg.ReturnSlaveBusyCountRequest, diag_msg.ReturnSlaveBusyCountResponse),
        (diag_msg.ReturnSlaveBusCharacterOverrunCountRequest, diag_msg.ReturnSlaveBusCharacterOverrunCountResponse),
        (diag_msg.ReturnIopOverrunCountRequest, diag_msg.ReturnIopOverrunCountResponse),
        (diag_msg.ClearOverrunCountRequest, diag_msg.ClearOverrunCountResponse),
        (diag_msg.GetClearModbusPlusRequest, diag_msg.GetClearModbusPlusResponse),
        (mei_msg.ReadDeviceInformationRequest, mei_msg.ReadDeviceInformationResponse),
    }

    def __init__(self, is_server: bool) -> None:
        """Initialize function_tables."""
        inx = 0 if is_server else 1
        self.lookup: dict[int, type[base.ModbusPDU]] = {cl[inx].function_code: cl[inx] for cl in self._pdu_class_table}
        self.sub_lookup: dict[int, dict[int, type[base.ModbusPDU]]] = {}
        for f in self._pdu_sub_class_table:
            if (function_code := f[inx].function_code) not in self.sub_lookup:
                self.sub_lookup[function_code] = {f[inx].sub_function_code: f[inx]}
            else:
                self.sub_lookup[function_code][f[inx].sub_function_code] = f[inx]

    def lookupPduClass(self, data: bytes) -> type[base.ModbusPDU] | None:
        """Use `function_code` to determine the class of the PDU."""
        func_code = int(data[1])
        if func_code & 0x80:
            return base.ExceptionResponse
        if func_code == 0x2B:  # mei message, sub_function_code is 1 byte
            sub_func_code = int(data[2])
            return self.sub_lookup[func_code].get(sub_func_code, None)
        if func_code == 0x08:  # diag message,  sub_function_code is 2 bytes
            sub_func_code = int(data[3])
            return self.sub_lookup[func_code].get(sub_func_code, None)
        return self.lookup.get(func_code, None)

    def register(self, custom_class: type[base.ModbusPDU]) -> None:
        """Register a function and sub function class with the decoder."""
        if not issubclass(custom_class, base.ModbusPDU):
            raise MessageRegisterException(
                f'"{custom_class.__class__.__name__}" is Not a valid Modbus Message'
                ". Class needs to be derived from "
                "`pymodbus.pdu.ModbusPDU` "
            )
        self.lookup[custom_class.function_code] = custom_class
        if custom_class.sub_function_code >= 0:
            if custom_class.function_code not in self.sub_lookup:
                self.sub_lookup[custom_class.function_code] = {}
            self.sub_lookup[custom_class.function_code][
                custom_class.sub_function_code
            ] = custom_class

    def decode(self, frame: bytes) -> base.ModbusPDU | None:
        """Decode a frame."""
        try:
            if (function_code := int(frame[0])) > 0x80:
                pdu_exp = base.ExceptionResponse(function_code & 0x7F)
                pdu_exp.decode(frame[1:])
                return pdu_exp
            if not (pdu_type := self.lookup.get(function_code, None)):
                Log.debug("decode PDU failed for function code {}", function_code)
                raise ModbusException(f"Unknown response {function_code}")
            pdu = pdu_type()
            pdu.decode(frame[1:])
            Log.debug("decoded PDU function_code({} sub {}) -> {} ", pdu.function_code, pdu.sub_function_code, str(pdu))

            if pdu.sub_function_code >= 0:
                lookup = self.sub_lookup.get(pdu.function_code, {})
                if subtype := lookup.get(pdu.sub_function_code, None):
                    pdu.__class__ = subtype
            return pdu
        except (ModbusException, ValueError, IndexError) as exc:
            Log.warning("Unable to decode frame {}", exc)
        return None



"""Modbus client async serial communication."""
from __future__ import annotations

import contextlib
import sys
import time
from collections.abc import Callable
from functools import partial

from pymodbus.client.base import ModbusBaseClient, ModbusBaseSyncClient
from pymodbus.exceptions import ConnectionException
from pymodbus.framer import FramerType
from pymodbus.logging import Log
from pymodbus.pdu import ModbusPDU
from pymodbus.transport import CommParams, CommType


with contextlib.suppress(ImportError):
    import serial


class AsyncModbusSerialClient(ModbusBaseClient):
    """**AsyncModbusSerialClient**.

    Fixed parameters:

    :param port: Serial port used for communication.

    Optional parameters:

    :param framer: Framer name, default FramerType.RTU
    :param baudrate: Bits per second.
    :param bytesize: Number of bits per byte 7-8.
    :param parity: 'E'ven, 'O'dd or 'N'one
    :param stopbits: Number of stop bits 1, 1.5, 2.
    :param handle_local_echo: Discard local echo from dongle.
    :param name: Set communication name, used in logging
    :param reconnect_delay: Minimum delay in seconds.milliseconds before reconnecting.
    :param reconnect_delay_max: Maximum delay in seconds.milliseconds before reconnecting.
    :param timeout: Timeout for connecting and receiving data, in seconds.
    :param retries: Max number of retries per request.
    :param trace_packet: Called with bytestream received/to be sent
    :param trace_pdu: Called with PDU received/to be sent
    :param trace_connect: Called when connected/disconnected

    .. tip::
        The trace methods allow to modify the datastream/pdu !

    .. tip::
        **reconnect_delay** doubles automatically with each unsuccessful connect, from
        **reconnect_delay** to **reconnect_delay_max**.
        Set `reconnect_delay=0` to avoid automatic reconnection.

    Example::

        from pymodbus.client import AsyncModbusSerialClient

        async def run():
            client = AsyncModbusSerialClient("dev/serial0")

            await client.connect()
            ...
            client.close()

    Please refer to :ref:`Pymodbus internals` for advanced usage.
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        port: str,
        *,
        framer: FramerType = FramerType.RTU,
        baudrate: int = 19200,
        bytesize: int = 8,
        parity: str = "N",
        stopbits: int = 1,
        handle_local_echo: bool = False,
        name: str = "comm",
        reconnect_delay: float = 0.1,
        reconnect_delay_max: float = 300,
        timeout: float = 3,
        retries: int = 3,
        trace_packet: Callable[[bool, bytes], bytes] | None = None,
        trace_pdu: Callable[[bool, ModbusPDU], ModbusPDU] | None = None,
        trace_connect: Callable[[bool], None] | None = None,
    ) -> None:
        """Initialize Asyncio Modbus Serial Client."""
        if "serial" not in sys.modules:  # pragma: no cover
            raise RuntimeError(
                "Serial client requires pyserial "
                'Please install with "pip install pyserial" and try again.'
            )
        if framer not in [FramerType.ASCII, FramerType.RTU]:
            raise TypeError("Only FramerType RTU/ASCII allowed.")
        self.comm_params = CommParams(
            comm_type=CommType.SERIAL,
            host=port,
            baudrate=baudrate,
            bytesize=bytesize,
            parity=parity,
            stopbits=stopbits,
            handle_local_echo=handle_local_echo,
            comm_name=name,
            reconnect_delay=reconnect_delay,
            reconnect_delay_max=reconnect_delay_max,
            timeout_connect=timeout,
        )
        ModbusBaseClient.__init__(
            self,
            framer,
            retries,
            self.comm_params,
            trace_packet,
            trace_pdu,
            trace_connect,
        )


class ModbusSerialClient(ModbusBaseSyncClient):
    """**ModbusSerialClient**.

    Fixed parameters:

    :param port: Serial port used for communication.

    Optional parameters:

    :param framer: Framer name, default FramerType.RTU
    :param baudrate: Bits per second.
    :param bytesize: Number of bits per byte 7-8.
    :param parity: 'E'ven, 'O'dd or 'N'one
    :param stopbits: Number of stop bits 0-2.
    :param handle_local_echo: Discard local echo from dongle.
    :param name: Set communication name, used in logging
    :param reconnect_delay: Not used in the sync client
    :param reconnect_delay_max: Not used in the sync client
    :param timeout: Timeout for connecting and receiving data, in seconds.
    :param retries: Max number of retries per request.
    :param trace_packet: Called with bytestream received/to be sent
    :param trace_pdu: Called with PDU received/to be sent
    :param trace_connect: Called when connected/disconnected

    .. tip::
        The trace methods allow to modify the datastream/pdu !

    Example::

        from pymodbus.client import ModbusSerialClient

        def run():
            client = ModbusSerialClient("dev/serial0")

            client.connect()
            ...
            client.close()

    Please refer to :ref:`Pymodbus internals` for advanced usage.
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        port: str,
        *,
        framer: FramerType = FramerType.RTU,
        baudrate: int = 19200,
        bytesize: int = 8,
        parity: str = "N",
        stopbits: int = 1,
        handle_local_echo: bool = False,
        name: str = "comm",
        reconnect_delay: float = 0.1,
        reconnect_delay_max: float = 300,
        timeout: float = 3,
        retries: int = 3,
        trace_packet: Callable[[bool, bytes], bytes] | None = None,
        trace_pdu: Callable[[bool, ModbusPDU], ModbusPDU] | None = None,
        trace_connect: Callable[[bool], None] | None = None,
    ) -> None:
        """Initialize Modbus Serial Client."""
        if "serial" not in sys.modules:  # pragma: no cover
            raise RuntimeError(
                "Serial client requires pyserial "
                'Please install with "pip install pyserial" and try again.'
            )
        if framer not in [FramerType.ASCII, FramerType.RTU]:
            raise TypeError("Only RTU/ASCII allowed.")
        self.comm_params = CommParams(
            comm_type=CommType.SERIAL,
            host=port,
            baudrate=baudrate,
            bytesize=bytesize,
            parity=parity,
            stopbits=stopbits,
            handle_local_echo=handle_local_echo,
            comm_name=name,
            reconnect_delay=reconnect_delay,
            reconnect_delay_max=reconnect_delay_max,
            timeout_connect=timeout,
        )
        super().__init__(
            framer,
            retries,
            self.comm_params,
            trace_packet,
            trace_pdu,
            trace_connect,
        )
        self.socket: serial.Serial | None = None
        self.last_frame_end = None
        self._t0 = float(1 + bytesize + stopbits) / baudrate

        # Check every 4 bytes / 2 registers if the reading is ready
        self._recv_interval = self._t0 * 4
        # Set a minimum of 1ms for high baudrates
        self._recv_interval = max(self._recv_interval, 0.001)

        self.inter_byte_timeout: float = 0
        self.silent_interval: float = 0
        if baudrate > 19200:
            self.silent_interval = 1.75 / 1000  # ms
        else:
            self.inter_byte_timeout = 1.5 * self._t0
            self.silent_interval = 3.5 * self._t0
        self.silent_interval = round(self.silent_interval, 6)

    @property
    def connected(self) -> bool:
        """Check if socket exists."""
        return self.socket is not None

    def connect(self) -> bool:
        """Connect to the modbus serial server."""
        if self.socket:
            return True
        try:
            self.socket = serial.serial_for_url(
                self.comm_params.host,
                timeout=self.comm_params.timeout_connect,
                bytesize=self.comm_params.bytesize,
                stopbits=self.comm_params.stopbits,
                baudrate=self.comm_params.baudrate,
                parity=self.comm_params.parity,
                exclusive=True,
            )
            self.socket.inter_byte_timeout = self.inter_byte_timeout
            self.last_frame_end = None
        # except serial.SerialException as msg:
        # pyserial raises undocumented exceptions like termios
        except Exception as msg:  # pylint: disable=broad-exception-caught
            Log.error("{}", msg)
            self.close()
        return self.socket is not None

    def close(self):
        """Close the underlying socket connection."""
        if self.socket:
            self.socket.close()
        self.socket = None

    def _in_waiting(self):
        """Return waiting bytes."""
        return getattr(self.socket, "in_waiting") if hasattr(self.socket, "in_waiting") else getattr(self.socket, "inWaiting")()

    def send(self, request: bytes, addr: tuple | None = None) -> int:
        """Send data on the underlying socket."""
        _ = addr
        if not self.socket:
            raise ConnectionException(str(self))
        if request:
            if waitingbytes := self._in_waiting():
                result = self.socket.read(waitingbytes)
                Log.warning("Cleanup recv buffer before send: {}", result, ":hex")
            if (size := self.socket.write(request)) is None:
                size = 0
            return size
        return 0

    def _wait_for_data(self) -> int:
        """Wait for data."""
        size = 0
        more_data = False
        condition = partial(
            lambda start, timeout: (time.time() - start) <= timeout,
            timeout=self.comm_params.timeout_connect,
        )
        start = time.time()
        while condition(start):
            available = self._in_waiting()
            if (more_data and not available) or (more_data and available == size):
                break
            if available and available != size:
                more_data = True
                size = available
            time.sleep(self._recv_interval)
        return size

    def recv(self, size: int | None) -> bytes:
        """Read data from the underlying descriptor."""
        if not self.socket:
            raise ConnectionException(str(self))
        if size is None:
            size = self._wait_for_data()
        if size > self._in_waiting():
            self._wait_for_data()
        result = self.socket.read(size)
        self.last_frame_end = round(time.time(), 6)
        return result

    def is_socket_open(self) -> bool:
        """Check if socket is open."""
        if self.socket:
            return self.socket.is_open
        return False

    def __repr__(self):
        """Return string representation."""
        return (
            f"<{self.__class__.__name__} at {hex(id(self))} socket={self.socket}, "
            f"framer={self.framer}, timeout={self.comm_params.timeout_connect}>"
        )



import logging
from optparse import Values
from typing import Generator, Iterable, Iterator, List, NamedTuple, Optional

from pip._vendor.packaging.requirements import InvalidRequirement
from pip._vendor.packaging.utils import canonicalize_name

from pip._internal.cli.base_command import Command
from pip._internal.cli.status_codes import ERROR, SUCCESS
from pip._internal.metadata import BaseDistribution, get_default_environment
from pip._internal.utils.misc import write_output

logger = logging.getLogger(__name__)


class ShowCommand(Command):
    """
    Show information about one or more installed packages.

    The output is in RFC-compliant mail header format.
    """

    usage = """
      %prog [options] <package> ..."""
    ignore_require_venv = True

    def add_options(self) -> None:
        self.cmd_opts.add_option(
            "-f",
            "--files",
            dest="files",
            action="store_true",
            default=False,
            help="Show the full list of installed files for each package.",
        )

        self.parser.insert_option_group(0, self.cmd_opts)

    def run(self, options: Values, args: List[str]) -> int:
        if not args:
            logger.warning("ERROR: Please provide a package name or names.")
            return ERROR
        query = args

        results = search_packages_info(query)
        if not print_results(
            results, list_files=options.files, verbose=options.verbose
        ):
            return ERROR
        return SUCCESS


class _PackageInfo(NamedTuple):
    name: str
    version: str
    location: str
    editable_project_location: Optional[str]
    requires: List[str]
    required_by: List[str]
    installer: str
    metadata_version: str
    classifiers: List[str]
    summary: str
    homepage: str
    project_urls: List[str]
    author: str
    author_email: str
    license: str
    license_expression: str
    entry_points: List[str]
    files: Optional[List[str]]


def search_packages_info(query: List[str]) -> Generator[_PackageInfo, None, None]:
    """
    Gather details from installed distributions. Print distribution name,
    version, location, and installed files. Installed files requires a
    pip generated 'installed-files.txt' in the distributions '.egg-info'
    directory.
    """
    env = get_default_environment()

    installed = {dist.canonical_name: dist for dist in env.iter_all_distributions()}
    query_names = [canonicalize_name(name) for name in query]
    missing = sorted(
        [name for name, pkg in zip(query, query_names) if pkg not in installed]
    )
    if missing:
        logger.warning("Package(s) not found: %s", ", ".join(missing))

    def _get_requiring_packages(current_dist: BaseDistribution) -> Iterator[str]:
        return (
            dist.metadata["Name"] or "UNKNOWN"
            for dist in installed.values()
            if current_dist.canonical_name
            in {canonicalize_name(d.name) for d in dist.iter_dependencies()}
        )

    for query_name in query_names:
        try:
            dist = installed[query_name]
        except KeyError:
            continue

        try:
            requires = sorted(
                # Avoid duplicates in requirements (e.g. due to environment markers).
                {req.name for req in dist.iter_dependencies()},
                key=str.lower,
            )
        except InvalidRequirement:
            requires = sorted(dist.iter_raw_dependencies(), key=str.lower)

        try:
            required_by = sorted(_get_requiring_packages(dist), key=str.lower)
        except InvalidRequirement:
            required_by = ["#N/A"]

        try:
            entry_points_text = dist.read_text("entry_points.txt")
            entry_points = entry_points_text.splitlines(keepends=False)
        except FileNotFoundError:
            entry_points = []

        files_iter = dist.iter_declared_entries()
        if files_iter is None:
            files: Optional[List[str]] = None
        else:
            files = sorted(files_iter)

        metadata = dist.metadata

        project_urls = metadata.get_all("Project-URL", [])
        homepage = metadata.get("Home-page", "")
        if not homepage:
            # It's common that there is a "homepage" Project-URL, but Home-page
            # remains unset (especially as PEP 621 doesn't surface the field).
            #
            # This logic was taken from PyPI's codebase.
            for url in project_urls:
                url_label, url = url.split(",", maxsplit=1)
                normalized_label = (
                    url_label.casefold().replace("-", "").replace("_", "").strip()
                )
                if normalized_label == "homepage":
                    homepage = url.strip()
                    break

        yield _PackageInfo(
            name=dist.raw_name,
            version=dist.raw_version,
            location=dist.location or "",
            editable_project_location=dist.editable_project_location,
            requires=requires,
            required_by=required_by,
            installer=dist.installer,
            metadata_version=dist.metadata_version or "",
            classifiers=metadata.get_all("Classifier", []),
            summary=metadata.get("Summary", ""),
            homepage=homepage,
            project_urls=project_urls,
            author=metadata.get("Author", ""),
            author_email=metadata.get("Author-email", ""),
            license=metadata.get("License", ""),
            license_expression=metadata.get("License-Expression", ""),
            entry_points=entry_points,
            files=files,
        )


def print_results(
    distributions: Iterable[_PackageInfo],
    list_files: bool,
    verbose: bool,
) -> bool:
    """
    Print the information from installed distributions found.
    """
    results_printed = False
    for i, dist in enumerate(distributions):
        results_printed = True
        if i > 0:
            write_output("---")

        metadata_version_tuple = tuple(map(int, dist.metadata_version.split(".")))

        write_output("Name: %s", dist.name)
        write_output("Version: %s", dist.version)
        write_output("Summary: %s", dist.summary)
        write_output("Home-page: %s", dist.homepage)
        write_output("Author: %s", dist.author)
        write_output("Author-email: %s", dist.author_email)
        if metadata_version_tuple >= (2, 4) and dist.license_expression:
            write_output("License-Expression: %s", dist.license_expression)
        else:
            write_output("License: %s", dist.license)
        write_output("Location: %s", dist.location)
        if dist.editable_project_location is not None:
            write_output(
                "Editable project location: %s", dist.editable_project_location
            )
        write_output("Requires: %s", ", ".join(dist.requires))
        write_output("Required-by: %s", ", ".join(dist.required_by))

        if verbose:
            write_output("Metadata-Version: %s", dist.metadata_version)
            write_output("Installer: %s", dist.installer)
            write_output("Classifiers:")
            for classifier in dist.classifiers:
                write_output("  %s", classifier)
            write_output("Entry-points:")
            for entry in dist.entry_points:
                write_output("  %s", entry.strip())
            write_output("Project-URLs:")
            for project_url in dist.project_urls:
                write_output("  %s", project_url)
        if list_files:
            write_output("Files:")
            if dist.files is None:
                write_output("Cannot locate RECORD or installed-files.txt")
            else:
                for line in dist.files:
                    write_output("  %s", line.strip())
    return results_printed
